from typing import Dict

import numpy as np
import torch


kMaxStrLength = 70  # Max string length for fixed length byte encoding.


def convert_str_to_fixed_length_tensor(string: str, max_length: int = kMaxStrLength) -> torch.Tensor:
    """
    Converts a string into a fixed length tensor of type torch.uint8.

    Args:
        string (str): String to convert
        max_length (int): Fixed length, default kMaxStrLength

    Returns:
        torch.Tensor: A fixed length tensor of type torch.uint8 that encodes the string
    """
    assert (
        len(string) <= max_length
    ), f"Encountered string longer that maximum length supported ({max_length}): {string}"
    assert "\0" not in string, f"String contains 0 value used for padding: {string}"
    return torch.cat(
        (
            torch.ByteTensor(torch.ByteStorage.from_buffer(string.encode("ascii"))),  # type: ignore
            torch.zeros(max_length - len(string), dtype=torch.uint8),
        )
    )


def move_to_device(data: Dict[str, torch.Tensor], device: torch.device) -> Dict[str, torch.Tensor]:
    """Move the data dict to a given torch device

    :param data: the torch dict
    :param device: the device where to move each value of the dict
    :return: the torch dict on the new device
    """
    return {k: v.to(device) for k, v in data.items()}


def move_to_numpy(data: Dict[str, torch.Tensor]) -> Dict[str, np.ndarray]:
    """Move a torch dict into numpy (on cpu)

    :param data: the torch dict
    :return: the numpy dict
    """
    return {k: v.cpu().numpy() for k, v in data.items()}



from collections import defaultdict
from enum import IntEnum
from typing import Dict, List, Optional

import cv2
import numpy as np

from ..data.filter import filter_tl_faces_by_status
from ..data.map_api import InterpolationMethod, MapAPI, TLFacesColors
from ..geometry import rotation33_as_yaw, transform_point, transform_points
from .rasterizer import Rasterizer
from .render_context import RenderContext


# sub-pixel drawing precision constants
CV2_SUB_VALUES = {"shift": 9, "lineType": cv2.LINE_AA}
CV2_SHIFT_VALUE = 2 ** CV2_SUB_VALUES["shift"]
INTERPOLATION_POINTS = 20


class RasterEls(IntEnum):  # map elements
    LANE_NOTL = 0
    ROAD = 1
    CROSSWALK = 2


COLORS = {
    TLFacesColors.GREEN.name: (0, 255, 0),
    TLFacesColors.RED.name: (255, 0, 0),
    TLFacesColors.YELLOW.name: (255, 255, 0),
    RasterEls.LANE_NOTL.name: (255, 217, 82),
    RasterEls.ROAD.name: (17, 17, 31),
    RasterEls.CROSSWALK.name: (255, 117, 69),
}


def indices_in_bounds(center: np.ndarray, bounds: np.ndarray, half_extent: float) -> np.ndarray:
    """
    Get indices of elements for which the bounding box described by bounds intersects the one defined around
    center (square with side 2*half_side)

    Args:
        center (float): XY of the center
        bounds (np.ndarray): array of shape Nx2x2 [[x_min,y_min],[x_max, y_max]]
        half_extent (float): half the side of the bounding box centered around center

    Returns:
        np.ndarray: indices of elements inside radius from center
    """
    x_center, y_center = center

    x_min_in = x_center > bounds[:, 0, 0] - half_extent
    y_min_in = y_center > bounds[:, 0, 1] - half_extent
    x_max_in = x_center < bounds[:, 1, 0] + half_extent
    y_max_in = y_center < bounds[:, 1, 1] + half_extent
    return np.nonzero(x_min_in & y_min_in & x_max_in & y_max_in)[0]


def cv2_subpixel(coords: np.ndarray) -> np.ndarray:
    """
    Cast coordinates to numpy.int but keep fractional part by previously multiplying by 2**CV2_SHIFT
    cv2 calls will use shift to restore original values with higher precision

    Args:
        coords (np.ndarray): XY coords as float

    Returns:
        np.ndarray: XY coords as int for cv2 shift draw
    """
    coords = coords * CV2_SHIFT_VALUE
    coords = coords.astype(np.int)
    return coords


class SemanticRasterizer(Rasterizer):
    """
    Rasteriser for the vectorised semantic map (generally loaded from json files).
    """

    def __init__(
            self, render_context: RenderContext, semantic_map_path: str, world_to_ecef: np.ndarray,
    ):
        self.render_context = render_context
        self.raster_size = render_context.raster_size_px
        self.pixel_size = render_context.pixel_size_m
        self.ego_center = render_context.center_in_raster_ratio

        self.world_to_ecef = world_to_ecef

        self.mapAPI = MapAPI(semantic_map_path, world_to_ecef)

    def rasterize(
            self,
            history_frames: np.ndarray,
            history_agents: List[np.ndarray],
            history_tl_faces: List[np.ndarray],
            agent: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        if agent is None:
            ego_translation_m = history_frames[0]["ego_translation"]
            ego_yaw_rad = rotation33_as_yaw(history_frames[0]["ego_rotation"])
        else:
            ego_translation_m = np.append(agent["centroid"], history_frames[0]["ego_translation"][-1])
            ego_yaw_rad = agent["yaw"]

        raster_from_world = self.render_context.raster_from_world(ego_translation_m, ego_yaw_rad)
        world_from_raster = np.linalg.inv(raster_from_world)

        # get XY of center pixel in world coordinates
        center_in_raster_px = np.asarray(self.raster_size) * (0.5, 0.5)
        center_in_world_m = transform_point(center_in_raster_px, world_from_raster)

        sem_im = self.render_semantic_map(center_in_world_m, raster_from_world, history_tl_faces[0])
        return sem_im.astype(np.float32) / 255

    def render_semantic_map(
            self, center_in_world: np.ndarray, raster_from_world: np.ndarray, tl_faces: np.ndarray
    ) -> np.ndarray:
        """Renders the semantic map at given x,y coordinates.

        Args:
            center_in_world (np.ndarray): XY of the image center in world ref system
            raster_from_world (np.ndarray):
        Returns:
            np.ndarray: RGB raster

        """

        img = 255 * np.ones(shape=(self.raster_size[1], self.raster_size[0], 3), dtype=np.uint8)

        # filter using half a radius from the center
        raster_radius = float(np.linalg.norm(self.raster_size * self.pixel_size)) / 2

        # get active traffic light faces
        active_tl_ids = set(filter_tl_faces_by_status(tl_faces, "ACTIVE")["face_id"].tolist())

        # get all lanes as interpolation so that we can transform them all together

        lane_indices = indices_in_bounds(center_in_world, self.mapAPI.bounds_info["lanes"]["bounds"], raster_radius)
        lanes_mask: Dict[str, np.ndarray] = defaultdict(lambda: np.zeros(len(lane_indices) * 2, dtype=np.bool))
        lanes_area = np.zeros((len(lane_indices) * 2, INTERPOLATION_POINTS, 2))

        for idx, lane_idx in enumerate(lane_indices):
            lane_idx = self.mapAPI.bounds_info["lanes"]["ids"][lane_idx]

            # interpolate over polyline to always have the same number of points
            lane_coords = self.mapAPI.get_lane_as_interpolation(
                lane_idx, INTERPOLATION_POINTS, InterpolationMethod.INTER_ENSURE_LEN
            )
            lanes_area[idx * 2] = lane_coords["xyz_left"][:, :2]
            lanes_area[idx * 2 + 1] = lane_coords["xyz_right"][::-1, :2]

            lane_type = RasterEls.LANE_NOTL.name
            lane_tl_ids = set(self.mapAPI.get_lane_traffic_control_ids(lane_idx))
            for tl_id in lane_tl_ids.intersection(active_tl_ids):
                lane_type = self.mapAPI.get_color_for_face(tl_id)

            lanes_mask[lane_type][idx * 2: idx * 2 + 2] = True

        if len(lanes_area):
            lanes_area = cv2_subpixel(transform_points(lanes_area.reshape((-1, 2)), raster_from_world))

            for lane_area in lanes_area.reshape((-1, INTERPOLATION_POINTS * 2, 2)):
                # need to for-loop otherwise some of them are empty
                cv2.fillPoly(img, [lane_area], COLORS[RasterEls.ROAD.name], **CV2_SUB_VALUES)

            lanes_area = lanes_area.reshape((-1, INTERPOLATION_POINTS, 2))
            for name, mask in lanes_mask.items():  # draw each type of lane with its own color
                cv2.polylines(img, lanes_area[mask], False, COLORS[name], **CV2_SUB_VALUES)

        # plot crosswalks
        crosswalks = []
        for idx in indices_in_bounds(center_in_world, self.mapAPI.bounds_info["crosswalks"]["bounds"], raster_radius):
            crosswalk = self.mapAPI.get_crosswalk_coords(self.mapAPI.bounds_info["crosswalks"]["ids"][idx])
            xy_cross = cv2_subpixel(transform_points(crosswalk["xyz"][:, :2], raster_from_world))
            crosswalks.append(xy_cross)

        cv2.polylines(img, crosswalks, True, COLORS[RasterEls.CROSSWALK.name], **CV2_SUB_VALUES)

        return img

    def to_rgb(self, in_im: np.ndarray, **kwargs: dict) -> np.ndarray:
        return (in_im * 255).astype(np.uint8)

    def num_channels(self) -> int:
        return 3



from .box_rasterizer import BoxRasterizer
from .rasterizer import EGO_EXTENT_HEIGHT, EGO_EXTENT_LENGTH, EGO_EXTENT_WIDTH, Rasterizer
from .rasterizer_builder import build_rasterizer
from .render_context import RenderContext
from .sat_box_rasterizer import SatBoxRasterizer
from .satellite_image import get_sat_image_crop, get_sat_image_crop_scaled, get_sat_image_crop_scaled_from_ecef
from .satellite_rasterizer import SatelliteRasterizer
from .sem_box_rasterizer import SemBoxRasterizer
from .semantic_rasterizer import SemanticRasterizer
from .stub_rasterizer import StubRasterizer


__all__ = [
    "get_sat_image_crop_scaled_from_ecef",
    "get_sat_image_crop_scaled",
    "get_sat_image_crop",
    "Rasterizer",
    "RenderContext",
    "SatelliteRasterizer",
    "SemanticRasterizer",
    "StubRasterizer",
    "BoxRasterizer",
    "SatBoxRasterizer",
    "SemBoxRasterizer",
    "build_rasterizer",
    "EGO_EXTENT_WIDTH",
    "EGO_EXTENT_LENGTH",
    "EGO_EXTENT_HEIGHT",
]



from abc import ABC, abstractmethod
from typing import Dict, List, Optional

from l5kit.cle.metric_set import L5MetricSet
from l5kit.environment.gym_metric_set import L2DisplacementYawMetricSet
from l5kit.simulation.unroll import SimulationOutputCLE


class Reward(ABC):
    """Base class interface for gym environment reward."""
    #: The prefix that will identify this reward class
    reward_prefix: str

    @abstractmethod
    def reset(self) -> None:
        """Reset the reward state when new episode starts.
        """
        raise NotImplementedError

    @abstractmethod
    def get_reward(self, frame_index: int, simulated_outputs: List[SimulationOutputCLE]) -> Dict[str, float]:
        """Return the reward at a particular time-step during the episode.

        :param frame_index: the frame index for which reward is to be calculated
        :param simulated_outputs: the object contain the ego target and prediction attributes
        :return: reward at a particular frame index (time-step) during the episode containing total reward
            and individual components that make up the reward.
        """
        raise NotImplementedError


class L2DisplacementYawReward(Reward):
    """This class is responsible for calculating a reward based on
    (1) L2 displacement error on the (x, y) coordinates
    (2) Closest angle error on the yaw coordinate
    during close loop simulation within the gym-compatible L5Kit environment.

    :param reward_prefix: the prefix that will identify this reward class
    :param metric_set: the set of metrics to compute
    :param enable_clip: flag to determine whether to clip reward
    :param rew_clip_thresh: the threshold to clip the reward
    :param use_yaw: flag to penalize the yaw prediction
    :param yaw_weight: weight of the yaw error
    """

    def __init__(self, reward_prefix: str = "L2DisplacementYaw", metric_set: Optional[L5MetricSet] = None,
                 enable_clip: bool = True, rew_clip_thresh: float = 15.0,
                 use_yaw: Optional[bool] = True, yaw_weight: Optional[float] = 20.0) -> None:
        """Constructor method
        """
        self.reward_prefix = reward_prefix
        # Metric Set
        self.metric_set = metric_set if metric_set is not None else L2DisplacementYawMetricSet()

        # Verify that error metrics necessary for reward calculation are present in the metric set
        if 'yaw_error_closest_angle' not in self.metric_set.evaluation_plan.metrics_dict():
            raise RuntimeError('\'yaw_error_closest_angle\' missing in metric set')
        if 'displacement_error_l2' not in self.metric_set.evaluation_plan.metrics_dict():
            raise RuntimeError('\'displacement_error_l2\' missing in metric set')

        self.use_yaw = use_yaw
        self.yaw_weight = yaw_weight

        self.enable_clip = enable_clip
        self.rew_clip_thresh = rew_clip_thresh

    def reset(self) -> None:
        """Reset the closed loop evaluator when a new episode starts.
        """
        self.metric_set.reset()

    @staticmethod
    def slice_simulated_output(index: int, simulated_outputs: List[SimulationOutputCLE]) -> List[SimulationOutputCLE]:
        """ Slice the simulated output at a particular frame index.
        This prevent calculating metric over all frames.

        :param index: the frame index at which the simulation outputs is to be sliced
        :param simulated_outputs: the object contain the ego target and prediction attributes
        :return: the sliced simulation output
        """
        # Only the simulated and recorded ego states are used for metric calculation
        simulated_outputs[0].recorded_ego_states = simulated_outputs[0].recorded_ego_states[index:index + 1]
        simulated_outputs[0].simulated_ego_states = simulated_outputs[0].simulated_ego_states[index:index + 1]
        return simulated_outputs

    def get_reward(self, frame_index: int, simulated_outputs: List[SimulationOutputCLE]) -> Dict[str, float]:
        """Get the reward for the given step in close loop training.

        :param frame_index: the frame index for which reward is to be calculated
        :param simulated_outputs: the object contain the ego target and prediction attributes
        :return: the dictionary containing total reward and individual components that make up the reward
        """
        scene_id = simulated_outputs[0].scene_id

        # Get the simulated output value at frame index + 1
        simulated_outputs = self.slice_simulated_output(frame_index + 1, simulated_outputs)

        # Evaluate metrics on the sliced simulated output
        self.metric_set.evaluate(simulated_outputs)
        scene_metrics = self.metric_set.evaluator.scene_metric_results[scene_id]
        dist_error = scene_metrics['displacement_error_l2']
        yaw_error = self.yaw_weight * scene_metrics['yaw_error_closest_angle']

        # clip the distance error (in x, y) only, not the yaw error (yaw error is bounded).
        dist_reward = float(-dist_error.item())
        if self.enable_clip:
            dist_reward = max(-self.rew_clip_thresh, -dist_error.item())

        # use yaw
        yaw_reward = 0.0
        if self.use_yaw:
            yaw_reward -= yaw_error.item()

        # Total reward
        total_reward = dist_reward + yaw_reward

        reward_dict = {"total": total_reward, "distance": dist_reward, "yaw": yaw_reward}
        return reward_dict



from .combine import get_combined_scenes
from .filter import (filter_agents_by_frames, filter_agents_by_labels, filter_agents_by_track_id,
                     filter_tl_faces_by_frames, filter_tl_faces_by_status, get_agents_slice_from_frames,
                     get_frames_slice_from_scenes, get_tl_faces_slice_from_frames)
from .labels import PERCEPTION_LABEL_TO_INDEX, PERCEPTION_LABELS, TL_FACE_LABEL_TO_INDEX, TL_FACE_LABELS
from .local_data_manager import DataManager, LocalDataManager
from .map_api import MapAPI
from .zarr_dataset import AGENT_DTYPE, ChunkedDataset, FRAME_DTYPE, SCENE_DTYPE, TL_FACE_DTYPE
from .zarr_utils import zarr_concat


__all__ = [
    "get_combined_scenes",
    "DataManager",
    "LocalDataManager",
    "ChunkedDataset",
    "SCENE_DTYPE",
    "FRAME_DTYPE",
    "AGENT_DTYPE",
    "TL_FACE_DTYPE",
    "PERCEPTION_LABELS",
    "PERCEPTION_LABEL_TO_INDEX",
    "filter_agents_by_labels",
    "filter_agents_by_frames",
    "filter_tl_faces_by_frames",
    "MapAPI",
    "zarr_concat",
    "TL_FACE_LABEL_TO_INDEX",
    "TL_FACE_LABELS",
    "filter_tl_faces_by_status",
    "get_frames_slice_from_scenes",
    "get_tl_faces_slice_from_frames",
    "get_agents_slice_from_frames",
    "filter_agents_by_track_id",
]



import cv2
import numpy as np


def crop_rectangle_from_image(image: np.ndarray, corners: np.ndarray) -> np.ndarray:
    """``crop_rectangle_from_image`` takes an image and 4 corners in pixel coordinates, it returns
    the sub-image inside that cropped area, rotated upright.

    Args:
        image (np.ndarray): image to crop from
        corners (np.ndarray): corners, array of shape (4,2)

    Returns:
        np.ndarray: crop from input containing the corners
    """
    rect = cv2.minAreaRect(np.array(corners[:, ::-1], dtype=np.float32))
    center, size, theta = rect
    center, size = tuple(center), tuple(np.int0(size))

    M = cv2.getRotationMatrix2D(center, theta, 1)
    dest = cv2.warpAffine(image, M, tuple((np.array(image.shape[:2]) * 2).astype(np.int64)))

    out = cv2.getRectSubPix(dest, size, center)
    return out



from .ackerman_perturbation import AckermanPerturbation
from .ackerman_steering_model import fit_ackerman_model_approximate, fit_ackerman_model_exact
from .perturbation import Perturbation


__all__ = ["fit_ackerman_model_approximate", "fit_ackerman_model_exact", "AckermanPerturbation", "Perturbation"]



from typing import List, Optional, Tuple

import numpy as np

from ..data import (filter_agents_by_labels, filter_tl_faces_by_frames, get_agents_slice_from_frames,
                    get_tl_faces_slice_from_frames)
from ..data.filter import filter_agents_by_frames, filter_agents_by_track_id
from ..geometry import angular_distance, compute_agent_pose, rotation33_as_yaw, transform_points
from ..kinematic import Perturbation
from ..rasterization import EGO_EXTENT_HEIGHT, EGO_EXTENT_LENGTH, EGO_EXTENT_WIDTH, Rasterizer, RenderContext
from .slicing import get_future_slice, get_history_slice


def get_agent_context(
        state_index: int,
        frames: np.ndarray,
        agents: np.ndarray,
        tl_faces: np.ndarray,
        history_num_frames: int,
        future_num_frames: int,
) -> Tuple[np.ndarray, np.ndarray, List[np.ndarray], List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
    """
    Slice zarr or numpy arrays to get the context around the agent onf interest (both in space and time)

    Args:
        state_index (int): frame index inside the scene
        frames (np.ndarray): frames from the scene
        agents (np.ndarray): agents from the scene
        tl_faces (np.ndarray): tl_faces from the scene
        history_num_frames (int): how many frames in the past to slice
        future_num_frames (int): how many frames in the future to slice

    Returns:
        Tuple[np.ndarray, np.ndarray, List[np.ndarray], List[np.ndarray], List[np.ndarray], List[np.ndarray]]
    """

    #  the history slice is ordered starting from the latest frame and goes backward in time., ex. slice(100, 91, -2)
    history_slice = get_history_slice(state_index, history_num_frames, 1, include_current_state=True)
    future_slice = get_future_slice(state_index, future_num_frames, 1)

    history_frames = frames[history_slice].copy()  # copy() required if the object is a np.ndarray
    future_frames = frames[future_slice].copy()

    sorted_frames = np.concatenate((history_frames[::-1], future_frames))  # from past to future

    # get agents (past and future)
    agent_slice = get_agents_slice_from_frames(sorted_frames[0], sorted_frames[-1])
    agents = agents[agent_slice].copy()
    # sync interval with the agents array
    history_frames["agent_index_interval"] -= agent_slice.start
    future_frames["agent_index_interval"] -= agent_slice.start
    history_agents = filter_agents_by_frames(history_frames, agents)
    future_agents = filter_agents_by_frames(future_frames, agents)

    # get traffic lights (past and future)
    tl_slice = get_tl_faces_slice_from_frames(sorted_frames[0], sorted_frames[-1])
    tl_faces = tl_faces[tl_slice].copy()
    # sync interval with the traffic light faces array
    history_frames["traffic_light_faces_index_interval"] -= tl_slice.start
    future_frames["traffic_light_faces_index_interval"] -= tl_slice.start
    history_tl_faces = filter_tl_faces_by_frames(history_frames, tl_faces)
    future_tl_faces = filter_tl_faces_by_frames(future_frames, tl_faces)

    return history_frames, future_frames, history_agents, future_agents, history_tl_faces, future_tl_faces


def compute_agent_velocity(
        history_positions_m: np.ndarray, future_positions_m: np.ndarray, step_time: float
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute estimated velocities by finite differentiation on future positions as(pos(T+t) - pos(T))/t.
    This simple approach gives less than 0.5% velocity difference
    compared to (pos(T+t) - pos(T-t))/2t on v1.1/sample.zarr.tar.

    Args:
        history_positions_m (np.ndarray): history XY positions in meters
        future_positions_m (np.ndarray): future XY positions in meters
        step_time (np.ndarray): length of a step in second

    Returns:
        Tuple[np.ndarray, np.ndarray]: history and future XY speeds

    """
    assert step_time > np.finfo(float).eps, f"step_time must be greater then eps, got {step_time}"

    # sorted from current to future frame
    # [future_num_frames, 2]
    future_positions_diff_m = np.concatenate((future_positions_m[:1], np.diff(future_positions_m, axis=0)))
    # [future_num_frames, 2]
    future_vels_mps = np.float32(future_positions_diff_m / step_time)

    # sorted from current to past frame, current position is included in history positions
    # [history_num_frames, 2]
    history_positions_diff_m = -np.diff(history_positions_m, axis=0)
    # [history_num_frames, 2]
    history_vels_mps = np.float32(history_positions_diff_m / step_time)

    return history_vels_mps, future_vels_mps


def get_relative_poses(
        num_frames: int,
        frames: np.ndarray,
        selected_track_id: Optional[int],
        agents: List[np.ndarray],
        agent_from_world: np.ndarray,
        current_agent_yaw: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Internal function that creates the targets and availability masks for deep prediction-type models.
    The futures/history offset (in meters) are computed. When no info is available (e.g. agent not in frame)
    a 0 is set in the availability array (1 otherwise).

    Note: return dtype is float32, even if the provided args are float64. Still, the transformation
    between space is performed in float64 to ensure precision

    Args:
        num_frames (int): number of offset we want in the future/history
        frames (np.ndarray): available frames. This may be less than num_frames
        selected_track_id (Optional[int]): agent track_id or AV (None)
        agents (List[np.ndarray]): list of agents arrays (same len of frames)
        agent_from_world (np.ndarray): local from world matrix
        current_agent_yaw (float): angle of the agent at timestep 0

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]: position offsets, angle offsets, extent, availabilities

    """
    # How much the coordinates differ from the current state in meters.
    positions_m = np.zeros((num_frames, 2), dtype=agent_from_world.dtype)
    yaws_rad = np.zeros((num_frames, 1), dtype=np.float32)
    extents_m = np.zeros((num_frames, 2), dtype=np.float32)
    availabilities = np.zeros((num_frames,), dtype=np.float32)

    for i, (frame, frame_agents) in enumerate(zip(frames, agents)):
        if selected_track_id is None:
            agent_centroid_m = frame["ego_translation"][:2]
            agent_yaw_rad = rotation33_as_yaw(frame["ego_rotation"])
            agent_extent = (EGO_EXTENT_LENGTH, EGO_EXTENT_WIDTH)
        else:
            # it's not guaranteed the target will be in every frame
            try:
                agent = filter_agents_by_track_id(frame_agents, selected_track_id)[0]
                agent_centroid_m = agent["centroid"]
                agent_yaw_rad = agent["yaw"]
                agent_extent = agent["extent"][:2]
            except IndexError:
                availabilities[i] = 0.0  # keep track of invalid futures/history
                continue

        positions_m[i] = agent_centroid_m
        yaws_rad[i] = agent_yaw_rad
        extents_m[i] = agent_extent
        availabilities[i] = 1.0

    # batch transform to speed up
    positions_m = transform_points(positions_m, agent_from_world) * availabilities[:, np.newaxis]
    yaws_rad = angular_distance(yaws_rad, current_agent_yaw) * availabilities[:, np.newaxis]
    return positions_m.astype(np.float32), yaws_rad, extents_m, availabilities


def generate_agent_sample(
        state_index: int,
        frames: np.ndarray,
        agents: np.ndarray,
        tl_faces: np.ndarray,
        selected_track_id: Optional[int],
        render_context: RenderContext,
        history_num_frames: int,
        future_num_frames: int,
        step_time: float,
        filter_agents_threshold: float,
        rasterizer: Optional[Rasterizer] = None,
        perturbation: Optional[Perturbation] = None,
) -> dict:
    """Generates the inputs and targets to train a deep prediction model. A deep prediction model takes as input
    the state of the world (here: an image we will call the "raster"), and outputs where that agent will be some
    seconds into the future.

    This function has a lot of arguments and is intended for internal use, you should try to use higher level classes
    and partials that use this function.

    Args:
        state_index (int): The anchor frame index, i.e. the "current" timestep in the scene
        frames (np.ndarray): The scene frames array, can be numpy array or a zarr array
        agents (np.ndarray): The full agents array, can be numpy array or a zarr array
        tl_faces (np.ndarray): The full traffic light faces array, can be numpy array or a zarr array
        selected_track_id (Optional[int]): Either None for AV, or the ID of an agent that you want to
        predict the future of. This agent is centered in the raster and the returned targets are derived from
        their future states.
        render_context (RenderContext): The context for rasterisation
        history_num_frames (int): Amount of history frames to draw into the rasters
        future_num_frames (int): Amount of history frames to draw into the rasters
        step_time (float): seconds between consecutive steps
        filter_agents_threshold (float): Value between 0 and 1 to use as cutoff value for agent filtering
        based on their probability of being a relevant agent
        rasterizer (Optional[Rasterizer]): Rasterizer of some sort that draws a map image
        perturbation (Optional[Perturbation]): Object that perturbs the input and targets, used
        to train models that can recover from slight divergence from training set data

    Raises:
        IndexError: An IndexError is returned if the specified ``selected_track_id`` is not present in the scene
        or was filtered by applying the ``filter_agent_threshold`` probability filtering.

    Returns:
        dict: a dict object with the raster array, the future offset coordinates (meters),
        the future yaw angular offset, the future_availability as a binary mask
    """
    (
        history_frames,
        future_frames,
        history_agents,
        future_agents,
        history_tl_faces,
        future_tl_faces,
    ) = get_agent_context(state_index, frames, agents, tl_faces, history_num_frames, future_num_frames, )

    if perturbation is not None and len(future_frames) == future_num_frames:
        history_frames, future_frames = perturbation.perturb(
            history_frames=history_frames, future_frames=future_frames
        )

    # State you want to predict the future of.
    cur_frame = history_frames[0]
    cur_agents = history_agents[0]

    if selected_track_id is None:
        agent_centroid_m = cur_frame["ego_translation"][:2]
        agent_yaw_rad = rotation33_as_yaw(cur_frame["ego_rotation"])
        agent_extent_m = np.asarray((EGO_EXTENT_LENGTH, EGO_EXTENT_WIDTH, EGO_EXTENT_HEIGHT))
        selected_agent = None
    else:
        # this will raise IndexError if the agent is not in the frame or under agent-threshold
        # this is a strict error, we cannot recover from this situation
        filtered_agents = filter_agents_by_labels(cur_agents, filter_agents_threshold)
        agent = filter_agents_by_track_id(filtered_agents, selected_track_id)[0]
        agent_centroid_m = agent["centroid"]
        agent_yaw_rad = float(agent["yaw"])
        agent_extent_m = agent["extent"]
        selected_agent = agent

    input_im = (
        None
        if not rasterizer
        else rasterizer.rasterize(history_frames, history_agents, history_tl_faces, selected_agent)
    )

    world_from_agent = compute_agent_pose(agent_centroid_m, agent_yaw_rad)
    agent_from_world = np.linalg.inv(world_from_agent)
    raster_from_world = render_context.raster_from_world(agent_centroid_m, agent_yaw_rad)

    future_positions_m, future_yaws_rad, future_extents, future_availabilities = get_relative_poses(
        future_num_frames, future_frames, selected_track_id, future_agents, agent_from_world, agent_yaw_rad,
    )
    # history_num_frames + 1 because it also includes the current frame
    history_positions_m, history_yaws_rad, history_extents, history_availabilities = get_relative_poses(
        history_num_frames + 1, history_frames, selected_track_id, history_agents, agent_from_world, agent_yaw_rad,
    )

    history_vels_mps, future_vels_mps = compute_agent_velocity(history_positions_m, future_positions_m, step_time)

    result = {
        "frame_index": state_index,
        "image": input_im,
        "target_positions": future_positions_m,
        "target_yaws": future_yaws_rad,
        "target_velocities": future_vels_mps,
        "target_availabilities": future_availabilities,
        "history_positions": history_positions_m,
        "history_yaws": history_yaws_rad,
        "history_velocities": history_vels_mps,
        "history_availabilities": history_availabilities,
        "world_to_image": raster_from_world,  # TODO deprecate
        "raster_from_agent": raster_from_world @ world_from_agent,
        "raster_from_world": raster_from_world,
        "agent_from_world": agent_from_world,
        "world_from_agent": world_from_agent,
        "centroid": agent_centroid_m,
        "yaw": agent_yaw_rad,
        "extent": agent_extent_m,
        "history_extents": history_extents,
        "future_extents": future_extents,
    }
    if len(history_vels_mps) > 0:
        # estimated current speed based on displacement between current frame at T and past frame at T-1
        result["curr_speed"] = np.linalg.norm(history_vels_mps[0])
    return result



from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np


# Ford Fusion dimensions TODO move somewhere else
EGO_EXTENT_WIDTH = 1.85
EGO_EXTENT_LENGTH = 4.87
EGO_EXTENT_HEIGHT = 1.8  # Height includes sensor


class Rasterizer(ABC):
    """Base class for something that takes a single state of the world, and outputs a (multi-channel) image.
    """

    def __init__(self) -> None:
        pass  # TODO are we sure we don't want at least the pixel information here?

    @abstractmethod
    def rasterize(
            self,
            history_frames: np.ndarray,
            history_agents: List[np.ndarray],
            history_tl_faces: List[np.ndarray],
            agent: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        pass

    @abstractmethod
    def to_rgb(self, in_im: np.ndarray, **kwargs: dict) -> np.ndarray:
        pass

    @abstractmethod
    def num_channels(self) -> int:
        pass



from collections import defaultdict
from enum import IntEnum
from typing import DefaultDict, Dict, List, NamedTuple, Optional, Set, Tuple

import numpy as np
import torch
from torch.utils.data.dataloader import default_collate
from tqdm.auto import tqdm

from l5kit.data import AGENT_DTYPE, PERCEPTION_LABEL_TO_INDEX
from l5kit.dataset import EgoDataset
from l5kit.dataset.utils import move_to_device, move_to_numpy
from l5kit.geometry import rotation33_as_yaw, transform_points
from l5kit.simulation.dataset import SimulationConfig, SimulationDataset


class TrajectoryStateIndices(IntEnum):
    """Defines indices for accessing trajectory states from LocalSceneBatch.
    example: all_speeds = local_scene_batch.recorded_ego_states[:, TrajectoryStateIndices.SPEED]

    :param X: the index for x position
    :param Y: the index for y position
    :param THETA: the index for the angle in radians
    :param SPEED: the index for speed in mps
    :param ACCELERATION: the index for acceleration in mps^2
    :param CURVATURE: the index for curvature (1pm)
    :param TIME: the index for time (seconds)
    """
    X = 0
    Y = 1
    THETA = 2
    SPEED = 3
    ACCELERATION = 4
    CURVATURE = 5
    TIME = 6


class UnrollInputOutput(NamedTuple):
    """ A single input output dict for an agent/ego in a given frame

    :param track_id: the agent track id
    :param inputs: input dict
    :param outputs: output dict
    """
    track_id: int
    inputs: Dict[str, np.ndarray]
    outputs: Dict[str, np.ndarray]


class ClosedLoopSimulatorModes(IntEnum):
    """Defines the different modes for which the closed loop simulator can be used.

    :param L5KIT: the index for using closed loop simulator for L5Kit environment
    :param GYM: the index for using closed loop simulator for Gym environment
    """
    L5KIT = 0
    GYM = 1


class SimulationOutputCLE:
    def __init__(self, scene_id: int, sim_dataset: SimulationDataset,
                 ego_ins_outs: DefaultDict[int, List[UnrollInputOutput]],
                 agents_ins_outs: DefaultDict[int, List[List[UnrollInputOutput]]]):
        """This object holds information about the result of the simulation loop
        for a given scene dataset for close loop evaluation.

        :param scene_id: the scene indices
        :param sim_dataset: the simulation dataset
        :param ego_ins_outs: all inputs and outputs for ego (each frame of each scene has only one)
        :param agents_ins_outs: all inputs and outputs for agents (multiple per frame in a scene)
        """
        if scene_id not in sim_dataset.scene_dataset_batch:
            raise ValueError(f"scene: {scene_id} not in sim datasets: {sim_dataset.scene_dataset_batch}")

        self.scene_id = scene_id
        recorded_dataset = sim_dataset.recorded_scene_dataset_batch[scene_id]
        simulated_dataset = sim_dataset.scene_dataset_batch[scene_id]

        self.simulated_agents = simulated_dataset.dataset.agents
        self.recorded_agents = recorded_dataset.dataset.agents
        self.recorded_ego = recorded_dataset.dataset.frames
        self.simulated_ego = simulated_dataset.dataset.frames

        self.simulated_ego_states = self.build_trajectory_states(self.simulated_ego)
        self.recorded_ego_states = self.build_trajectory_states(self.recorded_ego)

        self.ego_ins_outs = ego_ins_outs[scene_id]
        self.agents_ins_outs = agents_ins_outs[scene_id]

    def get_scene_id(self) -> int:
        """
        Get the scene index for this SimulationOutput

        :return: the scene index
        """
        return self.scene_id

    @staticmethod
    def build_trajectory_states(frames: np.ndarray) -> torch.Tensor:
        """
        Convert frames into a torch trajectory
        :param frames: the scene frames
        :return: the trajectory
        """
        trajectory_states = torch.zeros(len(frames), len(TrajectoryStateIndices), dtype=torch.float)
        translations = frames["ego_translation"]
        rotations = frames["ego_rotation"]

        for idx_frame in range(len(frames)):
            # TODO: there is a conversion from float64 to float32 here
            trajectory_states[idx_frame, TrajectoryStateIndices.X] = translations[idx_frame, 0]
            trajectory_states[idx_frame, TrajectoryStateIndices.Y] = translations[idx_frame, 1]
            trajectory_states[idx_frame, TrajectoryStateIndices.THETA] = rotation33_as_yaw(rotations[idx_frame])
            # TODO: we may need to fill other fields

        return trajectory_states


class SimulationOutput(SimulationOutputCLE):
    def __init__(self, scene_id: int, sim_dataset: SimulationDataset,
                 ego_ins_outs: DefaultDict[int, List[UnrollInputOutput]],
                 agents_ins_outs: DefaultDict[int, List[List[UnrollInputOutput]]]):
        """This object holds information about the result of the simulation loop
        for a given scene dataset for close loop evaluation and visualization

        :param scene_id: the scene indices
        :param sim_dataset: the simulation dataset
        :param ego_ins_outs: all inputs and outputs for ego (each frame of each scene has only one)
        :param agents_ins_outs: all inputs and outputs for agents (multiple per frame in a scene)
        """

        super(SimulationOutput, self).__init__(scene_id, sim_dataset, ego_ins_outs, agents_ins_outs)

        # Useful for visualization
        self.recorded_dataset = sim_dataset.recorded_scene_dataset_batch[scene_id]
        self.simulated_dataset = sim_dataset.scene_dataset_batch[scene_id]


class ClosedLoopSimulator:
    def __init__(self, sim_cfg: SimulationConfig, dataset: EgoDataset,
                 device: torch.device,
                 model_ego: Optional[torch.nn.Module] = None,
                 model_agents: Optional[torch.nn.Module] = None,
                 keys_to_exclude: Tuple[str] = ("image",),
                 mode: int = ClosedLoopSimulatorModes.L5KIT):
        """
        Create a simulation loop object capable of unrolling ego and agents
        :param sim_cfg: configuration for unroll
        :param dataset: EgoDataset used while unrolling
        :param device: a torch device. Inference will be performed here
        :param model_ego: the model to be used for ego
        :param model_agents: the model to be used for agents
        :param keys_to_exclude: keys to exclude from input/output (e.g. huge blobs)
        :param mode: the framework that uses the closed loop simulator
        """
        self.sim_cfg = sim_cfg
        if not sim_cfg.use_ego_gt and model_ego is None and mode == ClosedLoopSimulatorModes.L5KIT:
            raise ValueError("ego model should not be None when simulating ego")
        if not sim_cfg.use_agents_gt and model_agents is None and mode == ClosedLoopSimulatorModes.L5KIT:
            raise ValueError("agents model should not be None when simulating agent")
        if sim_cfg.use_ego_gt and mode == ClosedLoopSimulatorModes.GYM:
            raise ValueError("ego has to be simulated when using gym environment")
        if not sim_cfg.use_agents_gt and mode == ClosedLoopSimulatorModes.GYM:
            raise ValueError("agents need be log-replayed when using gym environment")

        self.model_ego = torch.nn.Sequential().to(device) if model_ego is None else model_ego.to(device)
        self.model_agents = torch.nn.Sequential().to(device) if model_agents is None else model_agents.to(device)

        self.device = device
        self.dataset = dataset

        self.keys_to_exclude = set(keys_to_exclude)

    def unroll(self, scene_indices: List[int]) -> List[SimulationOutput]:
        """
        Simulate the dataset for the given scene indices
        :param scene_indices: the scene indices we want to simulate
        :return: the simulated dataset
        """
        sim_dataset = SimulationDataset.from_dataset_indices(self.dataset, scene_indices, self.sim_cfg)

        agents_ins_outs: DefaultDict[int, List[List[UnrollInputOutput]]] = defaultdict(list)
        ego_ins_outs: DefaultDict[int, List[UnrollInputOutput]] = defaultdict(list)

        for frame_index in tqdm(range(len(sim_dataset)), disable=not self.sim_cfg.show_info):
            next_frame_index = frame_index + 1
            should_update = next_frame_index != len(sim_dataset)

            # AGENTS
            if not self.sim_cfg.use_agents_gt:
                agents_input = sim_dataset.rasterise_agents_frame_batch(frame_index)
                if len(agents_input):  # agents may not be available
                    agents_input_dict = default_collate(list(agents_input.values()))
                    agents_output_dict = self.model_agents(move_to_device(agents_input_dict, self.device))

                    # for update we need everything as numpy
                    agents_input_dict = move_to_numpy(agents_input_dict)
                    agents_output_dict = move_to_numpy(agents_output_dict)

                    if should_update:
                        self.update_agents(sim_dataset, next_frame_index, agents_input_dict, agents_output_dict)

                    # update input and output buffers
                    agents_frame_in_out = self.get_agents_in_out(agents_input_dict, agents_output_dict,
                                                                 self.keys_to_exclude)
                    for scene_idx in scene_indices:
                        agents_ins_outs[scene_idx].append(agents_frame_in_out.get(scene_idx, []))

            # EGO
            if not self.sim_cfg.use_ego_gt:
                ego_input = sim_dataset.rasterise_frame_batch(frame_index)
                ego_input_dict = default_collate(ego_input)
                ego_output_dict = self.model_ego(move_to_device(ego_input_dict, self.device))

                ego_input_dict = move_to_numpy(ego_input_dict)
                ego_output_dict = move_to_numpy(ego_output_dict)

                if should_update:
                    self.update_ego(sim_dataset, next_frame_index, ego_input_dict, ego_output_dict)

                ego_frame_in_out = self.get_ego_in_out(ego_input_dict, ego_output_dict, self.keys_to_exclude)
                for scene_idx in scene_indices:
                    ego_ins_outs[scene_idx].append(ego_frame_in_out[scene_idx])

        simulated_outputs: List[SimulationOutput] = []
        for scene_idx in scene_indices:
            simulated_outputs.append(SimulationOutput(scene_idx, sim_dataset, ego_ins_outs, agents_ins_outs))
        return simulated_outputs

    @staticmethod
    def update_agents(dataset: SimulationDataset, frame_idx: int, input_dict: Dict[str, np.ndarray],
                      output_dict: Dict[str, np.ndarray]) -> None:
        """Update the agents in frame_idx (across scenes) using agents_output_dict

        :param dataset: the simulation dataset
        :param frame_idx: index of the frame to modify
        :param input_dict: the input to the agent model
        :param output_dict: the output of the agent model
        :return:
        """

        agents_update_dict: Dict[Tuple[int, int], np.ndarray] = {}

        world_from_agent = input_dict["world_from_agent"]
        yaw = input_dict["yaw"]
        pred_trs = transform_points(output_dict["positions"][:, :1], world_from_agent)[:, 0]
        pred_yaws = yaw + output_dict["yaws"][:, 0, 0]

        next_agents = np.zeros(len(yaw), dtype=AGENT_DTYPE)
        next_agents["centroid"] = pred_trs
        next_agents["yaw"] = pred_yaws
        next_agents["track_id"] = input_dict["track_id"]
        next_agents["extent"] = input_dict["extent"]

        next_agents["label_probabilities"][:, PERCEPTION_LABEL_TO_INDEX["PERCEPTION_LABEL_CAR"]] = 1

        for scene_idx, next_agent in zip(input_dict["scene_index"], next_agents):
            agents_update_dict[(scene_idx, next_agent["track_id"])] = np.expand_dims(next_agent, 0)
        dataset.set_agents(frame_idx, agents_update_dict)

    @staticmethod
    def get_agents_in_out(input_dict: Dict[str, np.ndarray],
                          output_dict: Dict[str, np.ndarray],
                          keys_to_exclude: Optional[Set[str]] = None) -> Dict[int, List[UnrollInputOutput]]:
        """Get all agents inputs and outputs as a dict mapping scene index to a list of UnrollInputOutput

        :param input_dict: all agent model inputs (across scenes)
        :param output_dict: all agent model outputs (across scenes)
        :param keys_to_exclude: if to drop keys from input/output (e.g. huge blobs)
        :return: the dict mapping scene index to a list UnrollInputOutput. Some scenes may be missing
        """
        key_required = {"track_id", "scene_index"}
        if len(key_required.intersection(input_dict.keys())) != len(key_required):
            raise ValueError(f"track_id and scene_index not found in keys {input_dict.keys()}")

        keys_to_exclude = keys_to_exclude if keys_to_exclude is not None else set()
        if len(key_required.intersection(keys_to_exclude)) != 0:
            raise ValueError(f"can't drop required keys: {keys_to_exclude}")

        ret_dict = defaultdict(list)
        for idx_agent in range(len(input_dict["track_id"])):
            agent_in = {k: v[idx_agent] for k, v in input_dict.items() if k not in keys_to_exclude}
            agent_out = {k: v[idx_agent] for k, v in output_dict.items() if k not in keys_to_exclude}
            ret_dict[agent_in["scene_index"]].append(UnrollInputOutput(track_id=agent_in["track_id"],
                                                                       inputs=agent_in,
                                                                       outputs=agent_out))
        return dict(ret_dict)

    @staticmethod
    def get_ego_in_out(input_dict: Dict[str, np.ndarray],
                       output_dict: Dict[str, np.ndarray],
                       keys_to_exclude: Optional[Set[str]] = None) -> Dict[int, UnrollInputOutput]:
        """Get all ego inputs and outputs as a dict mapping scene index to a single UnrollInputOutput

        :param input_dict: all ego model inputs (across scenes)
        :param output_dict: all ego model outputs (across scenes)
        :param keys_to_exclude: if to drop keys from input/output (e.g. huge blobs)
        :return: the dict mapping scene index to a single UnrollInputOutput.
        """
        key_required = {"track_id", "scene_index"}
        if len(key_required.intersection(input_dict.keys())) != len(key_required):
            raise ValueError(f"track_id and scene_index not found in keys {input_dict.keys()}")

        keys_to_exclude = keys_to_exclude if keys_to_exclude is not None else set()
        if len(key_required.intersection(keys_to_exclude)) != 0:
            raise ValueError(f"can't drop required keys: {keys_to_exclude}")

        ret_dict = {}
        scene_indices = input_dict["scene_index"]
        if len(np.unique(scene_indices)) != len(scene_indices):
            raise ValueError(f"repeated scene_index for ego! {scene_indices}")

        for idx_ego in range(len(scene_indices)):
            ego_in = {k: v[idx_ego] for k, v in input_dict.items() if k not in keys_to_exclude}
            ego_out = {k: v[idx_ego] for k, v in output_dict.items() if k not in keys_to_exclude}
            ret_dict[ego_in["scene_index"]] = UnrollInputOutput(track_id=ego_in["track_id"],
                                                                inputs=ego_in,
                                                                outputs=ego_out)
        return ret_dict

    @staticmethod
    def update_ego(dataset: SimulationDataset, frame_idx: int, input_dict: Dict[str, np.ndarray],
                   output_dict: Dict[str, np.ndarray]) -> None:
        """Update ego across scenes for the given frame index.

        :param dataset: The simulation dataset
        :param frame_idx: index of the frame to modify
        :param input_dict: the input to the ego model
        :param output_dict: the output of the ego model
        :return:
        """
        world_from_agent = input_dict["world_from_agent"]
        yaw = input_dict["yaw"]
        pred_trs = transform_points(output_dict["positions"][:, :1], world_from_agent)
        pred_yaws = np.expand_dims(yaw, -1) + output_dict["yaws"][:, :1, 0]

        dataset.set_ego(frame_idx, 0, pred_trs, pred_yaws)



from .chop_dataset import create_chopped_dataset
from .csv_utils import read_gt_csv, read_pred_csv, write_gt_csv, write_pred_csv
from .extract_ground_truth import export_zarr_to_csv
from .extract_metrics import compute_metrics_csv


__all__ = [
    "export_zarr_to_csv",
    "compute_metrics_csv",
    "read_gt_csv",
    "read_pred_csv",
    "write_gt_csv",
    "write_pred_csv",
    "create_chopped_dataset",
]



from abc import abstractmethod
from collections import defaultdict
from enum import IntEnum
from typing import Any, DefaultDict, Dict, List, NamedTuple, Optional, Type

import torch
from typing_extensions import Protocol

from l5kit.cle import metrics
from l5kit.simulation.unroll import SimulationOutputCLE, TrajectoryStateIndices


class ValidatorOutput(NamedTuple):
    """Output from validators. Validators should return a boolean
    telling if the scene is valid or not and a list of failed
    frames."""
    is_valid_scene: bool
    failed_frames: List[int]


class SupportsMetricValidate(Protocol):
    """Protocol supporting the validation for metrics. The evaluation plan
    has two main components: metrics and validators. Metrics are completely
    independent, but validators not, as they depend on metrics, therefore
    the validator needs to carry a list of metrics it requires to compute,
    otherwise, the evaluation plan is not consistent, and this is checked
    by the evaluation plan.
    """
    validator_name: str
    requires_metric: List[str]

    @abstractmethod
    def validate(self, metric_results: Dict[str, torch.Tensor],
                 simulation_output: SimulationOutputCLE) -> ValidatorOutput:
        """Apply the validator on the metric results.

        :param metric_results: results from all computed metrics
        :param simulation_output: output from the closed-loop simulator
        :returns: True if validator passed, False otherwise
        """
        raise NotImplementedError


class DurationMode(IntEnum):
    """For more information about duration mode, see RangeValidator docs."""
    TOTAL = 0
    CONTINUOUS = 1


class RangeValidator(SupportsMetricValidate):
    """Validates a metric based on specified range. The range validator will
    check: min_value < metric_value < max_value. It will also check if the
    metric violated a maximum duration (see duration_mode parameter notes).

    :param validator_name: name of the validator
    :param metric: metric class used to validate
    :param min_value: minimum value allowed (inclusive)
    :param max_value: maximum value allowed (inclusive)
    :param violation_duration_s: maximum allowed duration in seconds
                                 where the metric can be violated
    :param duration_mode: the duration mode can be "total" or "continuous". In
                          "total" mode, all violations are summed per scene, and
                          if they exceed the violation_duration_s parameter, then
                          it will not pass the validation. In "continuous" mode,
                          the violation_duration_s parameter must be exceded
                          on a continuous violation of the metric in time (without
                          a gap of non-violation). This parameter is ignored
                          if violation_duration_s is zero.
    """

    def __init__(self, validator_name: str, metric: Type[metrics.SupportsMetricCompute],
                 min_value: Optional[float] = None, max_value: Optional[float] = None,
                 violation_duration_s: float = 0.0, duration_mode: DurationMode = DurationMode.TOTAL):
        # Requires at least one value specification
        if min_value is None and max_value is None:
            raise ValueError("At least one parameter must be "
                             "specified: min_value or max_value.")

        if min_value is not None and max_value is not None:
            if min_value >= max_value:
                raise ValueError("Minimum value cannot be greater or equal"
                                 " to the maximum value.")

        self.validator_name = validator_name
        self.metric_name = metric.metric_name
        self.requires_metric = [self.metric_name]
        self.min_value = min_value
        self.max_value = max_value
        self.validation_duration_s = violation_duration_s
        self.duration_mode = duration_mode

    @staticmethod
    def cumsum_with_reset(timestamp_diff: torch.Tensor,
                          validation_mask: torch.Tensor) -> torch.Tensor:
        """Cumulative sum (cumsum) that takes into consideration a valid mask.
        If the valid mask is False on a timestamp, it will reset the accumulator.

        :param timestamp_diff: timestamps differentiated (1D array)
        :param validation_mask: a boolean mask with valid/invalid timestamps (1D array)
        :return: cumulative sum (1D array)
        """
        cumsum = torch.zeros_like(timestamp_diff)
        accumulator = 0.0
        for idx, (ts, vmask) in enumerate(zip(timestamp_diff, validation_mask)):
            if not vmask:
                accumulator = 0.0
            else:
                accumulator += ts.item()
            cumsum[idx] = accumulator
        return cumsum

    def validate(self, metric_results: Dict[str, torch.Tensor],
                 simulation_output: SimulationOutputCLE) -> ValidatorOutput:
        """Apply the validator on the results of the metric computation.

        :param metric_results: all metric results
        :param simulation_output: output from the closed-loop simulator
        :returns: True if validator passed, False otherwise
        """
        result = metric_results[self.metric_name]
        validation_mask = torch.zeros_like(result, dtype=torch.bool)

        if self.min_value is not None:
            validation_mask |= result < self.min_value

        if self.max_value is not None:
            validation_mask |= result > self.max_value

        # Immediate failure if there is a violation and the allowed
        # duration is zero
        if self.validation_duration_s <= 0.0:
            # Log the failed frames into the scene tracker
            failed_frame_indexes = torch.nonzero(validation_mask).squeeze(1)

            failed_frame_indexes = failed_frame_indexes.cpu().numpy().tolist()

            is_valid_scene = len(failed_frame_indexes) == 0
            return ValidatorOutput(is_valid_scene, failed_frame_indexes)

        # If duration is greater than zero, then we check
        # if there was a violation greater than the
        # allowed duration
        ego_states = simulation_output.simulated_ego_states
        timestamps = ego_states[:, TrajectoryStateIndices.TIME.value]

        # Diff of the timestamps
        pad = torch.as_tensor([0], device=timestamps.device)
        pad_ts = torch.cat((pad, timestamps))
        ts_diff = pad_ts[1:] - pad_ts[:-1]

        # Total mode: we sum all violation durations
        if self.duration_mode == DurationMode.TOTAL:
            # Build a cumulative sum (masked by the validation mask)
            ts_valid_cumsum = (ts_diff * validation_mask).cumsum(dim=0)
            ts_valid_cumsum = ts_valid_cumsum * validation_mask

        # Continuous mode: we check if any of the durations
        # violated the constraint
        if self.duration_mode == DurationMode.CONTINUOUS:
            # Cumulative sum here is computed with reset, if there is a valid
            # metric computation between two consecutive chunks of invalid
            # metrics, then this valid frame will reset the cumulative sum
            # of the timestamp diffs
            ts_valid_cumsum = RangeValidator.cumsum_with_reset(ts_diff, validation_mask)

        # Check which timestamps violated the duration
        ts_cumsum_violated = ts_valid_cumsum > self.validation_duration_s

        # Get the frame indexes and track them
        violation_indexes = torch.nonzero(ts_cumsum_violated).squeeze(1)
        violation_indexes = violation_indexes.cpu().numpy().tolist()

        is_valid_scene = len(violation_indexes) == 0
        return ValidatorOutput(is_valid_scene, violation_indexes)


class SupportsValidationAggregation(Protocol):
    """Protocol supporting the validation aggregation. This aggregator
    is responsible for aggregating results from the the validators and
    also doing a reduction step across multiple distributed nodes.
    """

    def aggregate(self, scene_validation_results:
                  Dict[int, Dict[str, ValidatorOutput]]) -> Dict[str, Any]:
        """This method will aggregate scenes locally and then will
        do the reduction step to aggregate data across distributed
        nodes.

        :param scene_validation_results: results from validator
                                         outputs per scene
        :returns: any result (it can be a composite object with scenes and frames
                  or just a float value) indexed by the validator name.
        """
        raise NotImplementedError


class ValidationCountingAggregator(SupportsValidationAggregation):
    """This aggregator will count (sum) the amount of invalid scenes or
    optionally the amount of failed frames on each scene.

    :param failed_frames: if True, it will count the number of frames
                          failed instead of scenes failed.
    """

    def __init__(self, failed_frames: bool = False):
        self.failed_frames = failed_frames

    def aggregate_scenes(self, scene_validation_results:
                         Dict[int, Dict[str, ValidatorOutput]]) -> Dict[str, Any]:
        """Aggregate the scenes locally on each node. This method will just
        sum the number of invalid scenes across all scenes in the node.

        :param scene_validation_results: results from validator
                                         outputs per scene
        :returns: a dictionary with the validation metric name as keys
                  and the sum of invalid scenes
        """
        aggregation: DefaultDict[str, int] = defaultdict(int)
        for _, validator_dict in scene_validation_results.items():
            for validator_name, validator_output in validator_dict.items():
                # Aggregate the number of failed frames in the scene
                if self.failed_frames:
                    aggregation[validator_name] += len(validator_output.failed_frames)
                else:  # or the number of scenes
                    aggregation[validator_name] += not validator_output.is_valid_scene
        aggregation_torch = {k: torch.as_tensor(v) for k, v in aggregation.items()}
        return aggregation_torch

    def aggregate(self, scene_validation_results:
                  Dict[int, Dict[str, ValidatorOutput]]) -> Dict[str, Any]:
        """This method will aggregate scenes locally.

        :param scene_validation_results: results from validator
                                         outputs per scene
        :returns: any result (it can be a composite object with scenes and frames
                  or just a float value) indexed by the validator name.
        """
        agg_scenes = self.aggregate_scenes(scene_validation_results)
        return agg_scenes


class FailedFrame(NamedTuple):
    """A named-tuple composed of the scene if and the
    frame index on that scene that caused the validator to fail."""
    scene_id: int
    frame_index: int


class ValidationFailedFramesAggregator:
    """This aggregator will aggregate all failed frames (and scenes)."""

    def aggregate_scenes(self, scene_validation_results:
                         Dict[int, Dict[str, ValidatorOutput]]) -> Dict[str, Any]:
        """This method will aggregate the failed scene/frame tuples locally
        at each node and then build a 1D torch tensor with the unpacked tuples such
        as [scene, frame, scene frame, (...)], to be able to gather them later on
        the reduction across different nodes.

        :param scene_validation_results: results from validator
                                         outputs per scene
        :returns: a dictionary, indexed by the validator name, with
                  FailedFrame list containing the scene/frames failed.
        """
        aggregation: DefaultDict[str, List[FailedFrame]] = defaultdict(list)

        for scene_id, validator_dict in scene_validation_results.items():
            for validator_name, validator_output in validator_dict.items():
                if len(validator_output.failed_frames) > 0:
                    failed_fames = [FailedFrame(scene_id, frame_index)
                                    for frame_index in validator_output.failed_frames]
                    aggregation[validator_name].extend(failed_fames)

        aggregation_torch = {k: torch.as_tensor(v) for k, v in aggregation.items()}
        return aggregation_torch

    def aggregate(self, scene_validation_results:
                  Dict[int, Dict[str, ValidatorOutput]]) -> Dict[str, Any]:
        """This method will aggregate scenes locally and then will
        do the reduction step to aggregate data across distributed
        nodes.

        :param scene_validation_results: results from validator
                                         outputs per scene
        :returns: any result (it can be a composite object with scenes and frames
                  or just a float value) indexed by the validator name.
        """
        agg_scenes = self.aggregate_scenes(scene_validation_results)
        return agg_scenes



from typing import Collection, Union

import numpy as np


def points_within_bounds(coords: np.ndarray, shape: Union[Collection[int], np.ndarray]) -> np.ndarray:
    """
    Arguments:
        coords (np.ndarray): (N,3)-shaped array containing points.
        shape (tuple of ints or np.ndarray): shape to use as bounds, should be length 3
    Returns:
        Binary mask for given coords array which is True for points that fall within the bounds of shape.
    """
    shape = np.array(shape)
    within_bounds = np.all(coords >= 0, axis=1) * np.all(coords < shape, axis=1)
    return within_bounds


def voxel_coords_to_intensity_grid(
        voxel_coords: np.ndarray, shape: tuple, dtype: np.dtype = np.float32, drop_out_of_bounds: bool = True
) -> np.ndarray:
    """ Puts coords into a grid: for each grid cell the number of points is written there.

    Arguments:
        voxel_coords (np.ndarray): input array with coords (N,3) in intensity grid
        shape (tuple of ints): intensity grid shape

    Keyword Arguments:
        dtype (data-type): data type for the intensity grid (default: {np.float32})
        drop_out_of_bounds (bool): [description] (default: {True})

    Returns:
        np.ndarray -- Array with given shape, the value of each cell is the amount of coords for that point.
    """
    if drop_out_of_bounds:
        voxel_coords = voxel_coords[points_within_bounds(voxel_coords, shape)]

    grid = np.zeros(shape, dtype=dtype)

    c, intensity = np.unique(voxel_coords, axis=0, return_counts=True)
    grid[c[:, 0], c[:, 1], c[:, 2]] = intensity

    return grid


def normalize_intensity(x: np.ndarray, max_intensity: float) -> np.ndarray:
    """Normalize (divide by max) and clip intensity values to fall between 0 and 1.

    Arguments:
        x (np.npdarray): numpy array of any shape
        max_intensity (float): Maximum intensity value (anything above this will become 1)

    Returns:
        np.ndarray -- array of same type and shape as x with values between 0 and 1 only
    """
    return (x / max_intensity).clip(0, 1)



from .agent_sampling import compute_agent_velocity, generate_agent_sample, get_agent_context, get_relative_poses
from .slicing import get_future_slice, get_history_slice


__all__ = [
    "get_history_slice",
    "get_future_slice",
    "generate_agent_sample",
    "get_agent_context",
    "get_relative_poses",
    "compute_agent_velocity",
]



from typing import Any, Optional, Tuple, Union

import cv2
import numpy as np

from l5kit.geometry import transform_point


def get_sat_image_crop_scaled_from_ecef(
        sat_image: np.ndarray,
        crop_size: Union[Tuple[int, int], np.ndarray],
        ecef_translation: np.ndarray,
        ecef_to_sat: np.ndarray,
        **kwargs: Any,
) -> np.ndarray:
    """Utility function, calls get_sat_image_crop_scaled, see that function for more details on additional
    keyword arguments (such as ``yaw``).

    Arguments:
        sat_image (np.ndarray): satellite image
        crop_size (Union[Tuple[int, int], np.ndarray]): size of desired crop in pixels
        ecef_translation (np.ndarray): 2D or 3D vector where to center the cropped image
        ecef_to_sat (np.ndarray): transform from ECEF to satellite image space

    Returns:
        np.ndarray -- a crop of satellite_image
    """
    sat_pixel_translation = transform_point(ecef_translation, ecef_to_sat)
    return get_sat_image_crop_scaled(sat_image, crop_size, sat_pixel_translation, **kwargs)


def get_sat_image_crop_scaled(
        sat_image: np.ndarray,
        crop_size: Union[Tuple[int, int], np.ndarray],
        sat_pixel_translation: np.ndarray,
        yaw: Optional[float] = None,
        sat_pixel_scale: float = 1.0,
        pixel_size: float = 1.0,
        interpolation: int = cv2.INTER_LINEAR,
) -> np.ndarray:
    """Calls `get_sat_image_crop` (see that function's docs for further details), and rescales taking
        into account a desired pixel size.

    Example:
        Desired ``crop_size`` is 200x200, and ``pixel_size`` is 0.5: we want an image that corresponds
        to 100x100 meters. This means it extracts a 33x33 image and scales it up to 200x200.

    Arguments:
        sat_image (np.ndarray): satellite image
        crop_size (Union[Tuple[int, int], np.ndarray]): size of desired crop in pixels
        sat_pixel_translation (np.ndarray): 2D or 3D vector where to center the cropped image in pixels.

    Keyword Arguments:
        yaw (Optional[float]): yaw in radians, 0 means no rotation is applied, which generally means up is North.
            default: {None})

        sat_pixel_scale (float): A `sat_pixel_scale` of 3.0 would means that every pixel in the sat
        image corresponds to 3m in the real world. (default: {1.0})
        pixel_size (float): [description] (default: {1.0})
        interpolation (int): [description] (default: {cv2.INTER_LINEAR})

    Returns:
        (np.ndarray): a crop of input ``sat_image``
    """

    max_crop_size = [max(crop_size), max(crop_size)]
    crop_size_in_meters = np.array(max_crop_size) * pixel_size
    crop_size_in_sat_pixels = np.int0(np.round(crop_size_in_meters / sat_pixel_scale))

    sat_crop = get_sat_image_crop(sat_image, crop_size_in_sat_pixels, sat_pixel_translation, yaw)

    resized_sat_crop = cv2.resize(sat_crop, tuple(max_crop_size), interpolation=interpolation)
    start_x = np.int0(resized_sat_crop.shape[0] / 2 - crop_size[1] / 2)
    end_x = start_x + crop_size[1]
    start_y = np.int0(resized_sat_crop.shape[1] / 2 - crop_size[0] / 2)
    end_y = start_y + crop_size[0]

    out_sat_crop = resized_sat_crop[start_x:end_x, start_y:end_y]

    return out_sat_crop


def get_sat_image_crop(
        sat_image: np.ndarray,
        crop_size: Union[Tuple[int, int], np.ndarray],
        sat_pixel_translation: np.ndarray,
        yaw: Optional[float] = None,
) -> np.ndarray:
    """Crops input satellite such that ``sat_pixel_translation`` is centered in the image.

    Arguments:
        sat_image (np.ndarray): satellite image
        crop_size (Union[Tuple[int, int], np.ndarray]): size of desired crop in pixels
        sat_pixel_translation (np.ndarray): 2D or 3D vector where to center the cropped image in pixels.

    Keyword Arguments:
        yaw (Optional[float]): yaw in radians, None or 0 means no rotation is applied to the output image.
            default: {None})

    Returns:
        (np.ndarray): a crop of input ``sat_image``
    """
    if yaw is None:
        return _get_sat_image_crop_without_rotation(sat_image, crop_size, sat_pixel_translation)

    # We scale the image up, so that when it is rotated we can take the center crop without
    # losing the corners
    crop_size_scaled = np.ceil(np.array(crop_size) * np.sqrt(2)).astype(np.int64)

    im = _get_sat_image_crop_without_rotation(sat_image, crop_size_scaled, sat_pixel_translation)

    rot_matrix = cv2.getRotationMatrix2D((im.shape[1] / 2, im.shape[0] / 2), angle=np.degrees(yaw), scale=1.0)
    rotated_cropped = cv2.warpAffine(im, rot_matrix, (im.shape[1], im.shape[0]))

    # Center crop
    start_x = np.int0(rotated_cropped.shape[0] / 2 - crop_size[0] / 2)
    end_x = start_x + crop_size[0]
    start_y = np.int0(rotated_cropped.shape[1] / 2 - crop_size[1] / 2)
    end_y = start_y + crop_size[1]

    return rotated_cropped[start_x:end_x, start_y:end_y]


def _get_sat_image_crop_without_rotation(
        sat_image: np.ndarray, crop_size: np.ndarray, sat_pixel_translation: np.ndarray
) -> np.ndarray:
    """
    Crops satellite image around given translation.

    sat_pixel_translation should be an array with shape (2,) or (3,).
    """
    assert len(crop_size) >= 2
    assert sat_pixel_translation.shape[0] >= 2

    start_x = np.int0(sat_pixel_translation[0] - crop_size[0] // 2)
    end_x = start_x + crop_size[0]
    start_y = np.int0(sat_pixel_translation[1] - crop_size[1] // 2)
    end_y = start_y + crop_size[1]

    # Crops outside of the image not supported.
    if start_x < 0 or start_y < 0 or end_x >= sat_image.shape[0] or end_y >= sat_image.shape[1]:
        raise IndexError(
            "Satellite image crop ({}-{}, {}-{}) is out of bounds for satellite image of shape {}".format(
                start_x, end_x, start_y, end_y, sat_image.shape
            )
        )

    return sat_image[start_x:end_x, start_y:end_y]



import math
from typing import Callable

import torch


# Error function type
ErrorFunction = Callable[[torch.Tensor, torch.Tensor], torch.Tensor]


def l2_error(pred: torch.Tensor, gt: torch.Tensor) -> torch.Tensor:
    """A function that takes pred, gt tensor and computes their L2 distance.

    :param pred: predicted tensor, size: [batch_size, num_dims]
    :param gt: gt tensor, size: [batch_size, num_dims]
    :return: l2 distance between the predicted and gt tensor, size: [batch_size,]
    """
    return torch.norm(pred - gt, p=2, dim=-1)


def closest_angle_error(angle_a: torch.Tensor, angle_b: torch.Tensor) -> torch.Tensor:
    """ Finds the closest angle between angle_b - angle_a in radians.

    :param angle_a: a Tensor of angles in radians
    :param angle_b: a Tensor of angles in radians
    :return: The relative angle error between A and B between [0, pi]
    """
    assert angle_a.shape == angle_b.shape
    two_pi = 2.0 * math.pi
    wrapped = torch.fmod(angle_b - angle_a, two_pi)
    closest_angle = torch.fmod(2.0 * wrapped, two_pi) - wrapped
    return torch.abs(closest_angle)



import warnings
from typing import Dict, List

import torch
import torch.nn as nn
from torchvision.models.resnet import resnet18, resnet50


class PlanningModel(nn.Module):
    def __init__(
            self,
            model_arch: str,
            num_input_channels: int,
            num_targets: int,
            weights_scaling: List[float],
            criterion: nn.Module,
            pretrained: bool = True,
    ) -> None:
        super().__init__()
        self.model_arch = model_arch
        self.num_input_channels = num_input_channels
        self.num_targets = num_targets
        self.register_buffer("weights_scaling", torch.tensor(weights_scaling))
        self.pretrained = pretrained
        self.criterion = criterion

        if pretrained and self.num_input_channels != 3:
            warnings.warn("There is no pre-trained model with num_in_channels != 3, first layer will be reset")

        if model_arch == "resnet18":
            self.model = resnet18(pretrained=pretrained)
            self.model.fc = nn.Linear(in_features=512, out_features=num_targets)
        elif model_arch == "resnet50":
            self.model = resnet50(pretrained=pretrained)
            self.model.fc = nn.Linear(in_features=2048, out_features=num_targets)
        else:
            raise NotImplementedError(f"Model arch {model_arch} unknown")

        if self.num_input_channels != 3:
            self.model.conv1 = nn.Conv2d(
                self.num_input_channels, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False
            )

    def forward(self, data_batch: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        # [batch_size, channels, height, width]
        image_batch = data_batch["image"]
        # [batch_size, num_steps * 2]
        outputs = self.model(image_batch)
        batch_size = len(data_batch["image"])

        if self.training:
            if self.criterion is None:
                raise NotImplementedError("Loss function is undefined.")

            # [batch_size, num_steps * 2]
            targets = (torch.cat((data_batch["target_positions"], data_batch["target_yaws"]), dim=2)).view(
                batch_size, -1
            )
            # [batch_size, num_steps]
            target_weights = (data_batch["target_availabilities"].unsqueeze(-1) * self.weights_scaling).view(
                batch_size, -1
            )
            loss = torch.mean(self.criterion(outputs, targets) * target_weights)
            train_dict = {"loss": loss}
            return train_dict
        else:
            predicted = outputs.view(batch_size, -1, 3)
            # [batch_size, num_steps, 2->(XY)]
            pred_positions = predicted[:, :, :2]
            # [batch_size, num_steps, 1->(yaw)]
            pred_yaws = predicted[:, :, 2:3]
            eval_dict = {"positions": pred_positions, "yaws": pred_yaws}
            return eval_dict



import json

import yaml


def load_config_data(path: str) -> dict:
    """Load a config data from a given path

    :param path: the path as a string
    :return: the config as a dict
    """
    with open(path) as f:
        cfg: dict = yaml.load(f, Loader=yaml.FullLoader)
    return cfg


def save_config_data(data: dict, path: str) -> None:
    """Save the config on the disk

    :param data: the config as a dict
    :param path: the path as a string
    """
    with open(path, "w") as f:
        yaml.dump(data, f)


def load_metadata(metadata_path: str) -> dict:
    """Load the json metadata file

    Args:
        metadata_path: path to the dataset metadata

    Returns:
        dict: metadata as a dict
    """
    with open(metadata_path, "r") as f:
        metadata: dict = json.load(f)
    return metadata



from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Set

import bokeh.io
import bokeh.plotting
import numpy as np
from bokeh.layouts import column, LayoutDOM
from bokeh.models import CustomJS, HoverTool, Slider
from bokeh.plotting import ColumnDataSource

from l5kit.visualization.visualizer.common import (AgentVisualization, CWVisualization, EgoVisualization,
                                                   FrameVisualization, LaneVisualization, TrajectoryVisualization)


def _visualization_list_to_dict(visualisation_list: List[Any], null_el: Any) -> Dict[str, Any]:
    """Convert a list of NamedTuple into a dict, where:
    - the NamedTuple fields are the dict keys;
    - the dict value are lists;

    :param visualisation_list: a list of NamedTuple
    :param null_el: an element to be used as null if the list is empty (it can crash visualisation)
    :return: a dict with the same information
    """
    visualisation_list = visualisation_list if len(visualisation_list) else [null_el]
    visualisation_dict: DefaultDict[str, Any] = defaultdict(list)

    keys_set: Set[str] = set(visualisation_list[0]._asdict().keys())
    for el in visualisation_list:
        for k, v in el._asdict().items():
            if k not in keys_set:
                raise ValueError("keys set is not consistent between elements in the list")
            visualisation_dict[k].append(v)
    return dict(visualisation_dict)


def visualize(scene_index: int, frames: List[FrameVisualization]) -> LayoutDOM:
    """Visualise a scene using Bokeh.

    :param scene_index: the index of the scene, used only as the title
    :param frames: a list of FrameVisualization objects (one per frame of the scene)
    """

    agent_hover = HoverTool(
        mode="mouse",
        names=["agents"],
        tooltips=[
            ("Type", "@agent_type"),
            ("Probability", "@prob{0.00}%"),
            ("Track id", "@track_id"),
        ],
    )
    out: List[Dict[str, ColumnDataSource]] = []

    trajectories_labels = np.unique([traj.legend_label for frame in frames for traj in frame.trajectories])

    for frame_idx, frame in enumerate(frames):
        # we need to ensure we have something otherwise js crashes
        ego_dict = _visualization_list_to_dict([frame.ego], EgoVisualization(xs=np.empty(0), ys=np.empty(0),
                                                                             color="black", center_x=0,
                                                                             center_y=0))

        agents_dict = _visualization_list_to_dict(frame.agents, AgentVisualization(xs=np.empty(0), ys=np.empty(0),
                                                                                   color="black", track_id=-2,
                                                                                   agent_type="", prob=0.))

        lanes_dict = _visualization_list_to_dict(frame.lanes, LaneVisualization(xs=np.empty(0), ys=np.empty(0),
                                                                                color="black"))

        crosswalk_dict = _visualization_list_to_dict(frame.crosswalks, CWVisualization(xs=np.empty(0), ys=np.empty(0),
                                                                                       color="black"))

        # for trajectory we extract the labels so that we can show them in the legend
        trajectory_dict: Dict[str, Dict[str, Any]] = {}
        for trajectory_label in trajectories_labels:
            trajectories = [el for el in frame.trajectories if el.legend_label == trajectory_label]
            trajectory_dict[trajectory_label] = _visualization_list_to_dict(trajectories,
                                                                            TrajectoryVisualization(xs=np.empty(0),
                                                                                                    ys=np.empty(0),
                                                                                                    color="black",
                                                                                                    legend_label="none",
                                                                                                    track_id=-2))

        frame_dict = dict(ego=ColumnDataSource(ego_dict), agents=ColumnDataSource(agents_dict),
                          lanes=ColumnDataSource(lanes_dict),
                          crosswalks=ColumnDataSource(crosswalk_dict))
        frame_dict.update({k: ColumnDataSource(v) for k, v in trajectory_dict.items()})

        out.append(frame_dict)

    f = bokeh.plotting.figure(
        title="Scene {}".format(scene_index),
        match_aspect=True,
        x_range=(out[0]["ego"].data["center_x"][0] - 50, out[0]["ego"].data["center_x"][0] + 50),
        y_range=(out[0]["ego"].data["center_y"][0] - 50, out[0]["ego"].data["center_y"][0] + 50),
        tools=["pan", "wheel_zoom", agent_hover, "save", "reset"],
        active_scroll="wheel_zoom",
    )

    f.xgrid.grid_line_color = None
    f.ygrid.grid_line_color = None

    f.patches(line_width=0, alpha=0.5, color="color", source=out[0]["lanes"])
    f.patches(line_width=0, alpha=0.5, color="#B5B50D", source=out[0]["crosswalks"])
    f.patches(line_width=2, color="#B53331", source=out[0]["ego"])
    f.patches(line_width=2, color="color", name="agents", source=out[0]["agents"])

    js_string = """
            sources["lanes"].data = frames[cb_obj.value]["lanes"].data;
            sources["crosswalks"].data = frames[cb_obj.value]["crosswalks"].data;
            sources["agents"].data = frames[cb_obj.value]["agents"].data;
            sources["ego"].data = frames[cb_obj.value]["ego"].data;

            var center_x = frames[cb_obj.value]["ego"].data["center_x"][0];
            var center_y = frames[cb_obj.value]["ego"].data["center_y"][0];

            figure.x_range.setv({"start": center_x-50, "end": center_x+50})
            figure.y_range.setv({"start": center_y-50, "end": center_y+50})

            sources["lanes"].change.emit();
            sources["crosswalks"].change.emit();
            sources["agents"].change.emit();
            sources["ego"].change.emit();
        """

    for trajectory_name in trajectories_labels:
        f.multi_line(alpha=0.8, line_width=3, source=out[0][trajectory_name], color="color",
                     legend_label=trajectory_name)
        js_string += f'sources["{trajectory_name}"].data = frames[cb_obj.value]["{trajectory_name}"].data;\n' \
                     f'sources["{trajectory_name}"].change.emit();\n'

    slider_callback = CustomJS(
        args=dict(figure=f, sources=out[0], frames=out),
        code=js_string,
    )

    slider = Slider(start=0, end=len(frames), value=0, step=1, title="frame")
    slider.js_on_change("value", slider_callback)

    f.legend.location = "top_left"
    f.legend.click_policy = "hide"

    layout = column(f, slider)
    return layout



from typing import Iterable, Tuple

import cv2
import numpy as np


def write_video(
        output_filepath: str,
        images: Iterable[np.ndarray],
        resolution: Tuple[int, int],
        fps: float = 24.0,
        codec: str = "FMP4",
        interpolation: int = cv2.INTER_CUBIC,
) -> None:
    """Writes input RGB images to given output video filepath using OpenCV. It resizes images if
    necessary using given interpolation (default=``cv2.INTER_CUBIC``).

    Note that as this function uses OpenCV, your image's color channels will be inverted (RGB -> BGR) prior to writing.

    Arguments:
        output_filepath (str): output filepath, generally this should end in .mp4 or .avi
        depending on the codec used.
        images (Iterable[np.ndarray]): a list or other iterable of images.
        resolution (Tuple[int, int]): video resolution, e.g. (512, 512), the input frames are resized to this.

    Keyword Arguments:
        fps (float): Frames per second (default: {24.0})
        codec (str): Codec to be used. Note that with X264 codec only certain resolutions may work (default: {"FMP4"})
        interpolation (int): Interpolation to be used when resizing (default: {cv2.INTER_CUBIC})
    """

    fourcc = cv2.VideoWriter_fourcc(*codec)
    vw = cv2.VideoWriter(output_filepath, fourcc, fps, resolution)

    for img in images:
        # Go from C,0,1 to 0,1,C ordering
        if len(img.shape) == 3 and img.shape[0] == 3:
            img = img.transpose(1, 2, 0)

        if img.shape[:2] != resolution:
            img = cv2.resize(img, resolution, interpolation=interpolation)
        img = img[..., ::-1]  # RGB -> BGR
        vw.write(img)

    vw.release()



from typing import List, Optional

import cv2
import numpy as np

from ..geometry import rotation33_as_yaw, transform_point
from .rasterizer import Rasterizer
from .render_context import RenderContext
from .satellite_image import get_sat_image_crop_scaled


class SatelliteRasterizer(Rasterizer):
    """This rasterizer takes a satellite image in its constructor and a transform from world coordinates to this image.
    When you call rasterize, it will return a crop around the agent of interest with the agent's forward vector
    pointing right for the current timestep.
    """

    def __init__(
            self,
            render_context: RenderContext,
            map_im: np.ndarray,
            world_to_aerial: np.ndarray,
            interpolation: int = cv2.INTER_LINEAR,
    ):
        """

        Arguments:
            render_context (RenderContext): Render Context
            map_im (np.ndarray): Satellite image to crop from.
            world_to_aerial (np.ndarray): Transform to go from map coordinates to satellite image pixel coordinates.
        """
        self.render_context = render_context
        self.raster_size = render_context.raster_size_px
        self.pixel_size = render_context.pixel_size_m
        self.map_im = map_im
        self.world_to_aerial = world_to_aerial
        self.interpolation = interpolation
        self.map_pixel_scale = (
            1 / np.linalg.norm(world_to_aerial[0, 0:3]) + 1 / np.linalg.norm(world_to_aerial[1, 0:3])
        ) / 2

    def rasterize(
            self,
            history_frames: np.ndarray,
            history_agents: List[np.ndarray],
            history_tl_faces: List[np.ndarray],
            agent: Optional[np.ndarray] = None,
    ) -> np.ndarray:

        # Note 1: it looks like we are assuming that yaw in ecef == yaw in sat image
        if agent is None:
            ego_translation_m = history_frames[0]["ego_translation"]
            ego_yaw_rad = rotation33_as_yaw(history_frames[0]["ego_rotation"])

        else:
            ego_translation_m = np.append(agent["centroid"], history_frames[0]["ego_translation"][-1])
            ego_yaw_rad = agent["yaw"]

        raster_from_world = self.render_context.raster_from_world(ego_translation_m, ego_yaw_rad)
        world_from_raster = np.linalg.inv(raster_from_world)

        # Transform raster center to satellite coordinates (consider also z here)
        center_in_raster_px = np.asarray(self.raster_size) * (0.5, 0.5)
        center_in_world_m = transform_point(center_in_raster_px, world_from_raster)
        center_in_aerial_px = transform_point(np.append(center_in_world_m, ego_translation_m[2]), self.world_to_aerial)

        # Note 2: there is a negation here, unknown why this is necessary.
        # My best guess is because Y is flipped, maybe we can do this more elegantly.
        sat_im = get_sat_image_crop_scaled(
            self.map_im,
            self.raster_size,
            center_in_aerial_px,
            yaw=-ego_yaw_rad,
            pixel_size=self.pixel_size,
            sat_pixel_scale=self.map_pixel_scale,
            interpolation=self.interpolation,
        )
        if not self.render_context.set_origin_to_bottom:
            sat_im = sat_im[::-1]
        return sat_im.astype(np.float32) / 255

    def to_rgb(self, in_im: np.ndarray, **kwargs: dict) -> np.ndarray:
        return (in_im * 255).astype(np.uint8)

    def num_channels(self) -> int:
        return 3



from typing import Optional

import numpy as np
from tqdm import tqdm

from l5kit.data import ChunkedDataset
from l5kit.dataset import AgentDataset
from l5kit.geometry import transform_points
from l5kit.rasterization import RenderContext, StubRasterizer

from .csv_utils import write_gt_csv


def export_zarr_to_csv(
        zarr_dataset: ChunkedDataset,
        csv_file_path: str,
        future_num_frames: int,
        filter_agents_threshold: float,
        step_time: float = 0.1,
        agents_mask: Optional[np.array] = None,
) -> None:
    """Produces a csv file containing the ground truth from a zarr file.

    Arguments:
        zarr_dataset (np.ndarray): The open zarr dataset.
        csv_file_path (str): File path to write a CSV to.
        future_num_frames (int): Amount of future displacements we want.
        filter_agents_threshold (float): Value between 0 and 1 to use as cutoff value for agent filtering
        agents_mask (Optional[np.array]): a boolean mask of shape (len(zarr_dataset.agents)) which will be used
        instead of computing the agents_mask
    """

    cfg = {
        "raster_params": {
            "pixel_size": np.asarray((0.25, 0.25)),
            "raster_size": (100, 100),
            "filter_agents_threshold": filter_agents_threshold,
            "disable_traffic_light_faces": True,
            "ego_center": np.asarray((0.5, 0.5)),
            "set_origin_to_bottom": True,
        },
        "model_params": {"history_num_frames": 0, "future_num_frames": future_num_frames, "step_time": step_time},
    }

    render_context = RenderContext(
        np.asarray(cfg["raster_params"]["raster_size"]),
        cfg["raster_params"]["pixel_size"],
        cfg["raster_params"]["ego_center"],
        cfg["raster_params"]["set_origin_to_bottom"],
    )
    rasterizer = StubRasterizer(render_context)
    dataset = AgentDataset(cfg=cfg, zarr_dataset=zarr_dataset, rasterizer=rasterizer, agents_mask=agents_mask)

    future_coords_offsets = []
    target_availabilities = []

    timestamps = []
    agent_ids = []

    for el in tqdm(dataset, desc="extracting GT"):  # type: ignore
        # convert agent coordinates to world offsets
        offsets = transform_points(el["target_positions"], el["world_from_agent"]) - el["centroid"][:2]
        future_coords_offsets.append(offsets)

        timestamps.append(el["timestamp"])
        agent_ids.append(el["track_id"])
        target_availabilities.append(el["target_availabilities"])

    write_gt_csv(
        csv_file_path,
        np.asarray(timestamps),
        np.asarray(agent_ids),
        np.asarray(future_coords_offsets),
        np.asarray(target_availabilities),
    )



from typing import List

from l5kit.cle import metric_set, metrics


class L2DisplacementYawMetricSet(metric_set.L5MetricSet):
    """This class is responsible for computing a set of metric parametrization
    for the L5Kit Gym-compatible environment. The metrics in this class comprise
    L2 displacement error for the (x, y) coordinates and closest-angle error for the yaw.

    :param metric_prefix: this is a prefix that will identify the metric set being used.
    """

    def __init__(self, metric_prefix: str = "L2DisplacementYaw") -> None:
        """Constructor method
        """
        super().__init__(metric_prefix)

    def build_metrics(self) -> List[metrics.SupportsMetricCompute]:
        """Returns a list of metrics that will be computed.
        """
        return [
            metrics.DisplacementErrorL2Metric(),
            metrics.YawErrorCAMetric()
        ]



from typing import List, Optional

import numpy as np

from .labels import PERCEPTION_LABEL_TO_INDEX, TL_FACE_LABEL_TO_INDEX


# Labels that belong to "agents" of some sort.
PERCEPTION_LABELS_TO_KEEP = [
    "PERCEPTION_LABEL_CAR",
    "PERCEPTION_LABEL_VAN",
    "PERCEPTION_LABEL_TRAM",
    "PERCEPTION_LABEL_BUS",
    "PERCEPTION_LABEL_TRUCK",
    "PERCEPTION_LABEL_EMERGENCY_VEHICLE",
    "PERCEPTION_LABEL_OTHER_VEHICLE",
    "PERCEPTION_LABEL_BICYCLE",
    "PERCEPTION_LABEL_MOTORCYCLE",
    "PERCEPTION_LABEL_CYCLIST",
    "PERCEPTION_LABEL_MOTORCYCLIST",
    "PERCEPTION_LABEL_PEDESTRIAN",
    "PERCEPTION_LABEL_ANIMAL",
]
PERCEPTION_LABEL_INDICES_TO_KEEP = [PERCEPTION_LABEL_TO_INDEX[label] for label in PERCEPTION_LABELS_TO_KEEP]


def _get_label_filter(label_probabilities: np.ndarray, threshold: float) -> np.array:
    """

    Arguments:
        label_probabilities (np.ndarray): Given the probabilities of all labels, returns a binary mask
        of those whose summed probability of the classes we are interested in is higher than the given threshold.

        This set of classes "we are interested in" is hardcoded for now.

    Keyword Arguments:
        threshold (float): probability threshold for filtering

    Returns:
        np.array -- A binary array which can be used to mask agents.
    """
    return np.sum(label_probabilities[:, PERCEPTION_LABEL_INDICES_TO_KEEP], axis=1) > threshold


def filter_agents_by_labels(agents: np.ndarray, threshold: float = 0.5) -> np.ndarray:
    """Filters an agents array, keeping those agents that meet the threshold.

    Arguments:
        agents (np.ndarray): Agents array

    Keyword Arguments:
        threshold (float): probability threshold for filtering (default: {0.5})

    Returns:
        np.ndarray -- A subset of input ``agents`` array.
    """
    label_indices = _get_label_filter(agents["label_probabilities"], threshold)
    return agents[label_indices]


def filter_agents_by_track_id(agents: np.ndarray, track_id: int) -> np.ndarray:
    """Return all agent object (np.ndarray) of a given track_id.

    Arguments:
        agents (np.ndarray): agents array.
            NOTE: do NOT pass a zarr to this function, it can't handle boolean indexing
        track_id (int): agent track id to select

    Returns:
        np.ndarray -- Selected agent.
    """
    return agents[agents["track_id"] == track_id]


def filter_agents_by_frames(frames: np.ndarray, agents: np.ndarray) -> List[np.ndarray]:
    """
    Get a list of agents array, one array per frame. Note that "agent_index_interval" is used to filter agents,
    so you should take care of re-setting it if you have previously sliced agents.

    Args:
        frames (np.ndarray): an array of frames
        agents (np.ndarray): an array of agents

    Returns:
        List[np.ndarray] with the agents divided by frame
    """
    if frames.shape == ():
        frames = frames[None]  # add and axis if a single frame is passed
    return [agents[get_agents_slice_from_frames(frame)] for frame in frames]


def filter_tl_faces_by_frames(frames: np.ndarray, tl_faces: np.ndarray) -> List[np.ndarray]:
    """
    Get a list of traffic light faces arrays, one array per frame.
    This functions mimics `filter_agents_by_frames` for traffic light faces

    Args:
        frames (np.ndarray): an array of frames
        tl_faces (np.ndarray): an array of traffic light faces

    Returns:
        List[np.ndarray] with the traffic light faces divided by frame
    """
    return [tl_faces[get_tl_faces_slice_from_frames(frame)] for frame in frames]


def filter_tl_faces_by_status(tl_faces: np.ndarray, status: str) -> np.ndarray:
    """
    Filter tl_faces and keep only active faces
    Args:
        tl_faces (np.ndarray): array of traffic faces
        status (str): status we want to keep TODO refactor for enum

    Returns:
        np.ndarray: traffic light faces array with only faces with that status
    """
    return tl_faces[tl_faces["traffic_light_face_status"][:, TL_FACE_LABEL_TO_INDEX[status]] > 0]


def get_frames_slice_from_scenes(scene_a: np.ndarray, scene_b: Optional[np.ndarray] = None) -> slice:
    """
    Get a slice for indexing frames giving a start and end scene

    Args:
        scene_a (np.ndarray): the starting scene
        scene_b (Optional[np.ndarray]): the ending scene. If None, then scene_a end will be used

    Returns:
        slice: a slice object starting from the first frame in scene_a to the last one in scene_b
    """
    frame_index_start = scene_a["frame_index_interval"][0]
    if scene_b is None:
        scene_b = scene_a
    frame_index_end = scene_b["frame_index_interval"][1]
    return slice(frame_index_start, frame_index_end)


def get_agents_slice_from_frames(frame_a: np.ndarray, frame_b: Optional[np.ndarray] = None) -> slice:
    """
    Get a slice for indexing agents giving a start and end frame

    Args:
        frame_a (np.ndarray): the starting frame
        frame_b (Optional[np.ndarray]): the ending frame. If None, then frame_a end will be used

    Returns:
        slice: a slice object starting from the first agent in frame_a to the last one in frame_b
    """
    agent_index_start = frame_a["agent_index_interval"][0]
    if frame_b is None:
        frame_b = frame_a
    agent_index_end = frame_b["agent_index_interval"][1]
    return slice(agent_index_start, agent_index_end)


def get_tl_faces_slice_from_frames(frame_a: np.ndarray, frame_b: Optional[np.ndarray] = None) -> slice:
    """
    Get a slice for indexing traffic light faces giving a start and end frame

    Args:
        frame_a (np.ndarray): the starting frame
        frame_b (Optional[np.ndarray]): the ending frame. If None, then frame_a end will be used

    Returns:
        slice: a slice object starting from the first tl_face in frame_a to the last one in frame_b
    """
    tl_faces_index_start = frame_a["traffic_light_faces_index_interval"][0]
    if frame_b is None:
        frame_b = frame_a
    tl_faces_index_end = frame_b["traffic_light_faces_index_interval"][1]
    return slice(tl_faces_index_start, tl_faces_index_end)



from typing import cast, Sequence, Union

import numpy as np
import pymap3d as pm
import transforms3d


def compute_agent_pose(agent_centroid_m: np.ndarray, agent_yaw_rad: float) -> np.ndarray:
    """Return the agent pose as a 3x3 matrix. This corresponds to world_from_agent matrix.

    Args:
        agent_centroid_m (np.ndarry): 2D coordinates of the agent
        agent_yaw_rad (float): yaw of the agent

    Returns:
        (np.ndarray): 3x3 world_from_agent matrix
    """
    # Compute agent pose from its position and heading
    return np.array(
        [
            [np.cos(agent_yaw_rad), -np.sin(agent_yaw_rad), agent_centroid_m[0]],
            [np.sin(agent_yaw_rad), np.cos(agent_yaw_rad), agent_centroid_m[1]],
            [0, 0, 1],
        ]
    )


def rotation33_as_yaw(rotation: np.ndarray) -> float:
    """Compute the yaw component of given 3x3 rotation matrix.

    Args:
        rotation (np.ndarray): 3x3 rotation matrix (np.float64 dtype recommended)

    Returns:
        float: yaw rotation in radians
    """
    return cast(float, transforms3d.euler.mat2euler(rotation)[2])


def yaw_as_rotation33(yaw: float) -> np.ndarray:
    """Create a 3x3 rotation matrix from given yaw.
    The rotation is counter-clockwise and it is equivalent to:
    [cos(yaw), -sin(yaw), 0.0],
    [sin(yaw), cos(yaw), 0.0],
    [0.0, 0.0, 1.0],

    Args:
        yaw (float): yaw rotation in radians

    Returns:
        np.ndarray: 3x3 rotation matrix
    """
    return transforms3d.euler.euler2mat(0, 0, yaw)


def vertical_flip(tm: np.ndarray, y_dim_size: int) -> np.ndarray:
    """Return a new matrix that also performs a flip on the y axis.

    Args:
        tm: the original 3x3 matrix
        y_dim_size: this should match the resolution on y. It makes all coordinates positive

    Returns: a new 3x3 matrix.

    """
    flip_y = np.eye(3)
    flip_y[1, 1] = -1
    tm = np.matmul(flip_y, tm)
    tm[1, 2] += y_dim_size
    return tm


def transform_points(points: np.ndarray, transf_matrix: np.ndarray) -> np.ndarray:
    """
    Transform a set of 2D/3D points using the given transformation matrix.
    Assumes row major ordering of the input points. The transform function has 3 modes:
    - points (N, F), transf_matrix (F+1, F+1)
    all points are transformed using the matrix and the output points have shape (N, F).
    - points (B, N, F), transf_matrix (F+1, F+1)
    all sequences of points are transformed using the same matrix and the output points have shape (B, N, F).
    transf_matrix is broadcasted.
    - points (B, N, F), transf_matrix (B, F+1, F+1)
    each sequence of points is transformed using its own matrix and the output points have shape (B, N, F).
    Note this function assumes points.shape[-1] == matrix.shape[-1] - 1, which means that last
    rows in the matrices do not influence the final results.
    For 2D points only the first 2x3 parts of the matrices will be used.

    :param points: Input points of shape (N, F) or (B, N, F)
        with F = 2 or 3 depending on input points are 2D or 3D points.
    :param transf_matrix: Transformation matrix of shape (F+1, F+1) or (B, F+1, F+1) with F = 2 or 3.
    :return: Transformed points of shape (N, F) or (B, N, F) depending on the dimensions of the input points.
    """
    points_log = f" received points with shape {points.shape} "
    matrix_log = f" received matrices with shape {transf_matrix.shape} "

    assert points.ndim in [2, 3], f"points should have ndim in [2,3],{points_log}"
    assert transf_matrix.ndim in [2, 3], f"matrix should have ndim in [2,3],{matrix_log}"
    assert points.ndim >= transf_matrix.ndim, f"points ndim should be >= than matrix,{points_log},{matrix_log}"

    points_feat = points.shape[-1]
    assert points_feat in [2, 3], f"last points dimension must be 2 or 3,{points_log}"
    assert transf_matrix.shape[-1] == transf_matrix.shape[-2], f"matrix should be a square matrix,{matrix_log}"

    matrix_feat = transf_matrix.shape[-1]
    assert matrix_feat in [3, 4], f"last matrix dimension must be 3 or 4,{matrix_log}"
    assert points_feat == matrix_feat - 1, f"points last dim should be one less than matrix,{points_log},{matrix_log}"

    def _transform(points: np.ndarray, transf_matrix: np.ndarray) -> np.ndarray:
        num_dims = transf_matrix.shape[-1] - 1
        transf_matrix = np.transpose(transf_matrix, (0, 2, 1))
        return points @ transf_matrix[:, :num_dims, :num_dims] + transf_matrix[:, -1:, :num_dims]

    if points.ndim == transf_matrix.ndim == 2:
        points = np.expand_dims(points, 0)
        transf_matrix = np.expand_dims(transf_matrix, 0)
        return _transform(points, transf_matrix)[0]

    elif points.ndim == transf_matrix.ndim == 3:
        return _transform(points, transf_matrix)

    elif points.ndim == 3 and transf_matrix.ndim == 2:
        transf_matrix = np.expand_dims(transf_matrix, 0)
        return _transform(points, transf_matrix)
    else:
        raise NotImplementedError(f"unsupported case!{points_log},{matrix_log}")


def transform_point(point: np.ndarray, transf_matrix: np.ndarray) -> np.ndarray:
    """Transform a single vector using transformation matrix.
    This function call transform_points internally

    :param point: vector of shape (N)
    :param transf_matrix: transformation matrix of shape (N+1, N+1)
    :return: vector of same shape as input point
    """
    point = np.expand_dims(point, 0)
    return transform_points(point, transf_matrix)[0]


def ecef_to_geodetic(point: Union[np.ndarray, Sequence[float]]) -> np.ndarray:
    """Convert given ECEF coordinate into latitude, longitude, altitude.

    Args:
        point (Union[np.ndarray, Sequence[float]]): ECEF coordinate vector

    Returns:
        np.ndarray: latitude, altitude, longitude
    """
    return np.array(pm.ecef2geodetic(point[0], point[1], point[2]))


def geodetic_to_ecef(lla_point: Union[np.ndarray, Sequence[float]]) -> np.ndarray:
    """Convert given latitude, longitude, and optionally altitude into ECEF
    coordinates. If no altitude is given, altitude 0 is assumed.

    Args:
        lla_point (Union[np.ndarray, Sequence[float]]): Latitude, Longitude and optionally Altitude

    Returns:
        np.ndarray: 3D ECEF coordinate
    """
    if len(lla_point) == 2:
        return np.array(pm.geodetic2ecef(lla_point[0], lla_point[1], 0), dtype=np.float64)
    else:
        return np.array(pm.geodetic2ecef(lla_point[0], lla_point[1], lla_point[2]), dtype=np.float64)



from typing import List, NamedTuple

import numpy as np


class LaneVisualization(NamedTuple):
    """Hold information about a single lane

    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates
    :param color: color of the lane as a string (both hex or text)
    """
    xs: np.ndarray
    ys: np.ndarray
    color: str


class CWVisualization(NamedTuple):
    """Hold information about a single crosswalk

    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates
    :param color: color of the lane as a string (both hex or text)
    """
    xs: np.ndarray
    ys: np.ndarray
    color: str


class AgentVisualization(NamedTuple):
    """Hold information about a single agent

    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates
    :param color: color of the lane as a string (both hex or text)
    :param track_id: track id of the agent (unique in a scene)
    :param agent_type: type of the agent as a string (e.g. pedestrian)
    :param prob: probability of the agent from PCB
    """
    xs: np.ndarray
    ys: np.ndarray
    color: str
    track_id: int
    agent_type: str
    prob: float


class EgoVisualization(NamedTuple):
    """Hold information about a single ego annotation

    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates
    :param color: color of the lane as a string (both hex or text)
    :param center_x: the center x coordinate of the ego bbox
    :param center_y: the center y coordinate of the ego bbox
    """
    xs: np.ndarray
    ys: np.ndarray
    color: str
    center_x: float
    center_y: float


class TrajectoryVisualization(NamedTuple):
    """Hold information about a single trajectory annotation

    :param xs: 1D array of x coordinates
    :param ys: 1D array of y coordinates
    :param color: color of the lane as a string (both hex or text)
    :param legend_label: the name of this trajectory for the legend (e.g. `ego_trajectory`)
    :param track_id: the track id of the associated agent
    """
    xs: np.ndarray
    ys: np.ndarray
    color: str
    legend_label: str
    track_id: int


class FrameVisualization(NamedTuple):
    """Hold information about a frame (the state of a scene at a given time)

    :param ego: a single ego annotation
    :param agents: a list of agents
    :param lanes: a list of lanes
    :param crosswalks: a list of crosswalks
    :param trajectories: a list of trajectories
    """
    ego: EgoVisualization
    agents: List[AgentVisualization]
    lanes: List[LaneVisualization]
    crosswalks: List[CWVisualization]
    trajectories: List[TrajectoryVisualization]



from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple

import numpy as np

from l5kit.data import filter_agents_by_frames, PERCEPTION_LABEL_TO_INDEX
from l5kit.dataset import EgoDataset
from l5kit.geometry.transform import yaw_as_rotation33
from l5kit.simulation.utils import disable_agents, get_frames_subset, insert_agent


@dataclass
class SimulationConfig:
    """ Defines the parameters used for the simulation of ego and agents around it.

    :param use_ego_gt: whether to use GT annotations for ego instead of model's outputs
    :param use_agents_gt: whether to use GT annotations for agents instead of model's outputs
    :param disable_new_agents: whether to disable agents that are not returned at start_frame_index
    :param distance_th_far: if a tracked agent is closed than this value to ego, it will be controlled
    :param distance_th_close: if a new agent is closer than this value to ego, it will be controlled
    :param start_frame_index: the start index of the simulation
    :param num_simulation_steps: the number of step to simulate
    :param show_info: whether to show info logging during unroll
    """
    use_ego_gt: bool = False
    use_agents_gt: bool = False
    disable_new_agents: bool = False
    distance_th_far: float = 30.0
    distance_th_close: float = 15.0
    start_frame_index: int = 0
    num_simulation_steps: Optional[int] = None
    show_info: bool = False


class SimulationDataset:
    def __init__(self, scene_dataset_batch: Dict[int, EgoDataset], sim_cfg: SimulationConfig) -> None:
        """This class allows to:
        - rasterise the same frame across multiple scenes for ego;
        - rasterise the same frame across multiple scenes for multiple agents;
        - filter agents based on distance to ego;
        - set ego in future frames;
        - set agents in future frames;

        .. note:: only vehicles (car label) are picked as agents

        :param scene_dataset_batch: a mapping from scene index to EgoDataset
        :param sim_cfg: the simulation config
        """
        if not len(scene_dataset_batch):
            raise ValueError("can't build a simulation dataset with an empty batch")
        self.scene_dataset_batch: Dict[int, EgoDataset] = scene_dataset_batch
        self.sim_cfg = sim_cfg

        # we must limit the scenes to the part which will be simulated
        # we cut each scene so that it starts from there and ends after `num_simulation_steps`
        start_frame_idx = self.sim_cfg.start_frame_index
        if self.sim_cfg.num_simulation_steps is None:
            end_frame_idx = self.get_min_len()
        else:
            end_frame_idx = start_frame_idx + self.sim_cfg.num_simulation_steps
            if end_frame_idx > self.get_min_len():
                raise ValueError(f"can't unroll until frame {end_frame_idx}, length is {self.get_min_len()}")

        for scene_idx in scene_dataset_batch:
            zarr_dt = self.scene_dataset_batch[scene_idx].dataset
            self.scene_dataset_batch[scene_idx].dataset = get_frames_subset(zarr_dt, start_frame_idx, end_frame_idx)

            # this is the only stateful field we need to change for EgoDataset, it's used in bisect
            frame_index_ends = self.scene_dataset_batch[scene_idx].dataset.scenes["frame_index_interval"][:, 1]
            self.scene_dataset_batch[scene_idx].cumulative_sizes = frame_index_ends

        # buffer used to keep track of tracked agents during unroll as tuples of scene_idx, agent_idx
        self._agents_tracked: Set[Tuple[int, int]] = set()

        if self.sim_cfg.disable_new_agents:
            # we disable all agents that wouldn't be picked at frame 0
            for scene_idx, dt_ego in self.scene_dataset_batch.items():
                dataset_zarr = dt_ego.dataset
                frame = dataset_zarr.frames[0]
                ego_pos = frame["ego_translation"][:2]
                agents = dataset_zarr.agents
                frame_agents = filter_agents_by_frames(frame, agents)[0]
                frame_agents = self._filter_agents(scene_idx, frame_agents, ego_pos)
                disable_agents(dataset_zarr, allowlist=frame_agents["track_id"])

        # keep track of original dataset
        self.recorded_scene_dataset_batch = deepcopy(self.scene_dataset_batch)

    @staticmethod
    def from_dataset_indices(dataset: EgoDataset, scene_indices: List[int],
                             sim_cfg: SimulationConfig) -> "SimulationDataset":
        """Create a SimulationDataset by picking indices from the provided dataset

        :param dataset: the EgoDataset
        :param scene_indices: scenes from the EgoDataset to pick
        :param sim_cfg: a simulation config
        :return: the new SimulationDataset
        """
        if len(np.unique(scene_indices)) != len(scene_indices):
            raise ValueError(f"can't simulate repeated scenes: {scene_indices}")

        if np.any(np.asarray(scene_indices) >= len(dataset.dataset.scenes)):
            raise ValueError(
                f"can't pick indices {scene_indices} from dataset with length: {len(dataset.dataset.scenes)}")

        scene_dataset_batch: Dict[int, EgoDataset] = {}  # dicts preserve insertion order
        for scene_idx in scene_indices:
            scene_dataset = dataset.get_scene_dataset(scene_idx)
            scene_dataset_batch[scene_idx] = scene_dataset
        return SimulationDataset(scene_dataset_batch, sim_cfg)

    def get_min_len(self) -> int:
        """Return the minimum number of frames between the scenes

        :return: the minimum number of frames
        """
        return min([len(scene_dt.dataset.frames) for scene_dt in self.scene_dataset_batch.values()])

    def __len__(self) -> int:
        """
        Return the minimum number of frames across scenes

        :return: the number of frames
        """
        return self.get_min_len()

    def rasterise_frame_batch(self, state_index: int) -> List[Dict[str, np.ndarray]]:
        """
        Get a frame from all scenes

        :param state_index: the frame index
        :return: a list of dict from EgoDatasets
        """
        frame_batch = []
        for scene_idx, scene_dt in self.scene_dataset_batch.items():
            frame = scene_dt[state_index]
            frame["scene_index"] = scene_idx  # set the scene to the right index
            frame_batch.append(frame)
        return frame_batch

    def set_ego(self, state_index: int, output_index: int, ego_translations: np.ndarray,
                ego_yaws: np.ndarray) -> None:
        """Mutate future frame position and yaw for ego across scenes. This acts on the underlying dataset

        :param state_index: the frame index to mutate
        :param output_index: the index in ego_translations and ego_yaws to use
        :param ego_translations: output translations (N, T, 2)
        :param ego_yaws: output yaws (N, T)
        """

        if len(ego_translations) != len(ego_yaws):
            raise ValueError("lengths mismatch between translations and yaws")
        if len(ego_translations) != len(self.scene_dataset_batch):
            raise ValueError("lengths mismatch between scenes and predictions")
        if state_index >= len(self):
            raise ValueError(f"trying to mutate frame:{state_index} but length is:{len(self)}")

        position_m_batch = ego_translations[:, output_index, :]
        angle_rad_batch = ego_yaws[:, output_index]
        for i, (scene_dataset, position_m, angle_rad) in enumerate(
            zip(self.scene_dataset_batch.values(), position_m_batch, angle_rad_batch)
        ):
            scene_dataset.dataset.frames[state_index]["ego_translation"][:2] = position_m
            scene_dataset.dataset.frames[state_index]["ego_rotation"] = yaw_as_rotation33(angle_rad)

    def set_agents(self, state_index: int, agents_infos: Dict[Tuple[int, int], np.ndarray]) -> None:
        """Set multiple agents in the scene datasets.

        :param state_index: the frame index to set (same for all datasets)
        :param agents_infos: a dict mapping (scene_idx, agent_idx) to the agent array
        """
        for (scene_idx, _), agent in agents_infos.items():
            insert_agent(agent, state_index, self.scene_dataset_batch[scene_idx].dataset)

    def rasterise_agents_frame_batch(self, state_index: int) -> Dict[Tuple[int, int], Dict[str, np.ndarray]]:
        """Rasterise agents for each scene in the batch at a given frame.

        :param state_index: the frame index in the scene
        :return: a dict mapping from [scene_id, track_id] to the numpy dict
        """
        ret = {}
        for scene_index in self.scene_dataset_batch:
            ret.update(self._rasterise_agents_frame(scene_index, state_index))
        return ret

    def _rasterise_agents_frame(self, scene_index: int,
                                state_index: int) -> Dict[Tuple[int, int], Dict[str, np.ndarray]]:
        """Rasterise agents of interest for a given frame in a given scene.

        :param scene_index: index of the scene
        :param state_index: frame index
        :return: a dict mapping [scene_idx, agent_idx] to dict
        """
        # filter agents around ego based on distance and threshold
        dataset = self.scene_dataset_batch[scene_index]
        frame = dataset.dataset.frames[state_index]

        frame_agents = filter_agents_by_frames(frame, dataset.dataset.agents)[0]
        frame_agents = self._filter_agents(scene_index, frame_agents, frame["ego_translation"][:2])

        # rasterise individual agents
        agents_dict: Dict[Tuple[int, int], Dict[str, np.ndarray]] = {}
        for agent in frame_agents:
            track_id = int(agent["track_id"])
            el = dataset.get_frame(scene_index=0, state_index=state_index, track_id=track_id)
            # we replace the scene_index here to match the real one (otherwise is 0)
            el["scene_index"] = scene_index
            agents_dict[scene_index, track_id] = el

        self._update_agent_infos(scene_index, frame_agents["track_id"])
        return agents_dict

    def _update_agent_infos(self, scene_index: int, agent_track_ids: np.ndarray) -> None:
        """Update tracked agents object such that:
        - if agent was not there -> add it
        - if agent is not here anymore -> remove it
        This will be used next frame to control thresholds

        :param scene_index: index of the scene
        :param agent_track_ids: agents track ids for this frame
        """
        agent_track_set = set([(scene_index, int(track_id)) for track_id in agent_track_ids])

        self._agents_tracked.update(agent_track_set)

        remove_els = set([k for k in self._agents_tracked if k[0] == scene_index]) - agent_track_set
        for indices in remove_els:
            self._agents_tracked.remove(indices)

    def _filter_agents(self, scene_idx: int, frame_agents: np.ndarray,
                       ego_pos: np.ndarray) -> np.ndarray:
        """Filter agents according to a set of rules:
        if new agent (not in tracked_agents) then:
            - must be a car
            - must be in distance_th_close
        if tracked agent:
            - must be in distance_th_far

        This is to avoid acquiring and releasing the same agents if it is on the boundary of the selection

        :param scene_idx: the scene index (used to check for agents_infos)
        :param frame_agents: the agents in this frame
        :param ego_pos: the ego position in this frame
        :return: the filtered agents
        """
        # keep only vehicles
        car_index = PERCEPTION_LABEL_TO_INDEX["PERCEPTION_LABEL_CAR"]
        vehicle_mask = frame_agents["label_probabilities"][:, car_index]

        dt_agents_ths = self.scene_dataset_batch[scene_idx].cfg["raster_params"]["filter_agents_threshold"]
        vehicle_mask = vehicle_mask > dt_agents_ths
        frame_agents = frame_agents[vehicle_mask]

        distance_mask = np.zeros(len(frame_agents), dtype=np.bool)
        for idx_agent, agent in enumerate(frame_agents):
            track_id = int(agent["track_id"])

            distance = np.linalg.norm(ego_pos - agent["centroid"])
            # for distance use two thresholds
            if (scene_idx, track_id) in self._agents_tracked:
                # if we're already controlling this agent, th_far
                distance_mask[idx_agent] = distance < self.sim_cfg.distance_th_far
            else:
                # if not, start controlling it only if in th_close
                distance_mask[idx_agent] = distance < self.sim_cfg.distance_th_close

        return frame_agents[distance_mask]



from typing import List, Optional

import numpy as np

from .rasterizer import Rasterizer
from .render_context import RenderContext


class StubRasterizer(Rasterizer):
    """This rasterizer doesn't actually do anything, it returns an all-black image. Useful for testing.

    """

    def __init__(self, render_context: RenderContext):
        """

        Args:
            render_context (RenderContext): The render context used to convert coordinates into raster space
            filter_agents_threshold (float): Value between 0 and 1 used to filter uncertain agent detections
        """
        super(StubRasterizer, self).__init__()
        self.raster_size = render_context.raster_size_px
        self.pixel_size = render_context.pixel_size_m
        self.ego_center = render_context.center_in_raster_ratio

    def rasterize(
            self,
            history_frames: np.ndarray,
            history_agents: List[np.ndarray],
            history_tl_faces: List[np.ndarray],
            agent: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        """
        Just return an empty raster_size x 3 black image
        """
        out_im = np.zeros((self.raster_size[0], self.raster_size[1], 3), dtype=np.float32)
        return out_im

    def to_rgb(self, in_im: np.ndarray, **kwargs: dict) -> np.ndarray:
        """
        Return a completely black image.
        """
        return np.zeros((self.raster_size[0], self.raster_size[1], 3), dtype=np.uint8)

    def num_channels(self) -> int:
        return 3



from enum import IntEnum
from functools import lru_cache
from typing import Iterator, no_type_check, Sequence, Union

import numpy as np
import pymap3d as pm

from l5kit.configs.config import load_metadata
from l5kit.data import DataManager

from ..geometry import transform_points
from .proto.road_network_pb2 import GeoFrame, GlobalId, MapElement, MapFragment


CACHE_SIZE = int(1e5)
ENCODING = "utf-8"


class InterpolationMethod(IntEnum):
    INTER_METER = 0  # fixed interpolation at a given step in meters
    INTER_ENSURE_LEN = 1  # ensure we always get the same number of elements


class TLFacesColors(IntEnum):
    RED = 0
    GREEN = 1
    YELLOW = 2


class MapAPI:
    def __init__(self, protobuf_map_path: str, world_to_ecef: np.ndarray):
        """
        Interface to the raw protobuf map file with the following features:
        - access to element using ID is O(1);
        - access to coordinates in world ref system for a set of elements is O(1) after first access (lru cache)
        - object support iteration using __getitem__ protocol

        Args:
            protobuf_map_path (str): path to the protobuf file
            world_to_ecef (np.ndarray): transformation matrix from world coordinates to ECEF (dataset dependent)
        """
        self.protobuf_map_path = protobuf_map_path
        self.ecef_to_world = np.linalg.inv(world_to_ecef)

        with open(protobuf_map_path, "rb") as infile:
            mf = MapFragment()
            mf.ParseFromString(infile.read())

        self.elements = mf.elements
        self.ids_to_el = {self.id_as_str(el.id): idx for idx, el in enumerate(self.elements)}  # store a look-up table

        self.bounds_info = self.get_bounds()  # store bound for semantic elements for fast look-up

    @staticmethod
    def from_cfg(data_manager: DataManager, cfg: dict) -> "MapAPI":
        """Build a MapAPI object starting from a config file and a data manager

        :param data_manager: a data manager object ot resolve paths
        :param cfg: the config dict
        :return: a MapAPI object
        """
        raster_cfg = cfg["raster_params"]
        dataset_meta_key = raster_cfg["dataset_meta_key"]

        semantic_map_filepath = data_manager.require(raster_cfg["semantic_map_key"])
        dataset_meta = load_metadata(data_manager.require(dataset_meta_key))
        world_to_ecef = np.array(dataset_meta["world_to_ecef"], dtype=np.float64)

        return MapAPI(semantic_map_filepath, world_to_ecef)

    @staticmethod
    @no_type_check
    def id_as_str(element_id: GlobalId) -> str:
        """
        Get the element id as a string.
        Elements ids are stored as a variable len sequence of bytes in the protobuf

        Args:
            element_id (GlobalId): the GlobalId in the protobuf

        Returns:
            str: the id as a str
        """
        return element_id.id.decode(ENCODING)

    @staticmethod
    def _undo_e7(value: float) -> float:
        """
        Latitude and longitude are stored as value*1e7 in the protobuf for efficiency and guaranteed accuracy.
        Convert them back to float.

        Args:
            value (float): the scaled value

        Returns:
            float: the unscaled value
        """
        return value / 1e7

    @no_type_check
    def unpack_deltas_cm(self, dx: Sequence[int], dy: Sequence[int], dz: Sequence[int], frame: GeoFrame) -> np.ndarray:
        """
        Get coords in world reference system (local ENU->ECEF->world).
        See the protobuf annotations for additional information about how coordinates are stored

        Args:
            dx (Sequence[int]): X displacement in centimeters in local ENU
            dy (Sequence[int]): Y displacement in centimeters in local ENU
            dz (Sequence[int]): Z displacement in centimeters in local ENU
            frame (GeoFrame): geo-location information for the local ENU. It contains lat and long origin of the frame

        Returns:
            np.ndarray: array of shape (Nx3) with XYZ coordinates in world ref system

        """
        x = np.cumsum(np.asarray(dx) / 100)
        y = np.cumsum(np.asarray(dy) / 100)
        z = np.cumsum(np.asarray(dz) / 100)
        frame_lat, frame_lng = self._undo_e7(frame.origin.lat_e7), self._undo_e7(frame.origin.lng_e7)
        xyz = np.stack(pm.enu2ecef(x, y, z, frame_lat, frame_lng, 0), axis=-1)
        xyz = transform_points(xyz, self.ecef_to_world)
        return xyz

    @staticmethod
    @no_type_check
    def is_lane(element: MapElement) -> bool:
        """
        Check whether an element is a valid lane

        Args:
            element (MapElement): a proto element

        Returns:
            bool: True if the element is a valid lane
        """
        return bool(element.element.HasField("lane"))

    @lru_cache(maxsize=CACHE_SIZE)
    def get_lane_coords(self, element_id: str) -> dict:
        """
        Get XYZ coordinates in world ref system for a lane given its id
        lru_cached for O(1) access

        Args:
            element_id (str): lane element id

        Returns:
            dict: a dict with the two boundaries coordinates as (Nx3) XYZ arrays
        """
        element = self[element_id]
        assert self.is_lane(element)

        lane = element.element.lane
        left_boundary = lane.left_boundary
        right_boundary = lane.right_boundary

        xyz_left = self.unpack_deltas_cm(
            left_boundary.vertex_deltas_x_cm,
            left_boundary.vertex_deltas_y_cm,
            left_boundary.vertex_deltas_z_cm,
            lane.geo_frame,
        )
        xyz_right = self.unpack_deltas_cm(
            right_boundary.vertex_deltas_x_cm,
            right_boundary.vertex_deltas_y_cm,
            right_boundary.vertex_deltas_z_cm,
            lane.geo_frame,
        )

        return {"xyz_left": xyz_left, "xyz_right": xyz_right}

    @staticmethod
    def interpolate(xyz: np.ndarray, step: float, method: InterpolationMethod) -> np.ndarray:
        """
        Interpolate points based on cumulative distances from the first one. Two modes are available:
        INTER_METER: interpolate using step as a meter value over cumulative distances (variable len result)
        INTER_ENSURE_LEN: interpolate using a variable step such that we always get step values
        Args:
            xyz (np.ndarray): XYZ coords
            step (float): param for the interpolation
            method (InterpolationMethod): method to use to interpolate

        Returns:
            np.ndarray: the new interpolated coordinates
        """
        cum_dist = np.cumsum(np.linalg.norm(np.diff(xyz, axis=0), axis=-1))
        cum_dist = np.insert(cum_dist, 0, 0)

        if method == InterpolationMethod.INTER_ENSURE_LEN:
            step = int(step)
            assert step > 1, "step must be at least 2 with INTER_ENSURE_LEN"
            steps = np.linspace(cum_dist[0], cum_dist[-1], step)

        elif method == InterpolationMethod.INTER_METER:
            assert step > 0, "step must be greater than 0 with INTER_FIXED"
            steps = np.arange(cum_dist[0], cum_dist[-1], step)
        else:
            raise NotImplementedError(f"interpolation method should be in {InterpolationMethod.__members__}")

        xyz_inter = np.empty((len(steps), 3), dtype=xyz.dtype)
        xyz_inter[:, 0] = np.interp(steps, xp=cum_dist, fp=xyz[:, 0])
        xyz_inter[:, 1] = np.interp(steps, xp=cum_dist, fp=xyz[:, 1])
        xyz_inter[:, 2] = np.interp(steps, xp=cum_dist, fp=xyz[:, 2])
        return xyz_inter

    @lru_cache(maxsize=CACHE_SIZE)
    def get_lane_traffic_control_ids(self, element_id: str) -> set:
        lane = self[element_id].element.lane
        return set([MapAPI.id_as_str(la_tc) for la_tc in lane.traffic_controls])

    @lru_cache(maxsize=CACHE_SIZE)
    def get_lane_as_interpolation(self, element_id: str, step: float, method: InterpolationMethod) -> dict:
        """
        Perform an interpolation of the left and right lanes and compute the midlane.
        See interpolate for details about the different interpolation methods

        Args:
            element_id (str): lane id
            step (float): step param for the method
            method (InterpolationMethod): one of the accepted methods

        Returns:
            dict: same as `get_lane_coords` but overwrite xyz values for the lanes
        """
        lane_dict = self.get_lane_coords(element_id)
        xyz_left = lane_dict["xyz_left"]
        xyz_right = lane_dict["xyz_right"]

        lane_dict["xyz_left"] = self.interpolate(xyz_left, step, method)
        lane_dict["xyz_right"] = self.interpolate(xyz_right, step, method)

        # to compute midlane we average between left and right bounds
        # but to do that we need them to have the same numbers of points
        # if that's not the case (interpolation is not INTER_ENSURE_LEN) we interpolate again with that mode
        if method != InterpolationMethod.INTER_ENSURE_LEN:
            mid_steps = max(len(xyz_left), len(xyz_right))
            # recompute lanes using fixed length
            xyz_left = self.interpolate(xyz_left, mid_steps, InterpolationMethod.INTER_ENSURE_LEN)
            xyz_right = self.interpolate(xyz_right, mid_steps, InterpolationMethod.INTER_ENSURE_LEN)

        else:
            xyz_left = lane_dict["xyz_left"]
            xyz_right = lane_dict["xyz_right"]

        xyz_midlane = (xyz_left + xyz_right) / 2

        # interpolate xyz for midlane with the selected interpolation
        lane_dict["xyz_midlane"] = self.interpolate(xyz_midlane, step, method)
        return lane_dict

    @staticmethod
    @no_type_check
    def is_crosswalk(element: MapElement) -> bool:
        """
        Check whether an element is a valid crosswalk

        Args:
            element (MapElement): a proto element

        Returns:
            bool: True if the element is a valid crosswalk
        """
        if not element.element.HasField("traffic_control_element"):
            return False
        traffic_element = element.element.traffic_control_element
        return bool(traffic_element.HasField("pedestrian_crosswalk") and traffic_element.points_x_deltas_cm)

    @lru_cache(maxsize=CACHE_SIZE)
    def get_crosswalk_coords(self, element_id: str) -> dict:
        """
        Get XYZ coordinates in world ref system for a crosswalk given its id
        lru_cached for O(1) access

        Args:
            element_id (str): crosswalk element id

        Returns:
            dict: a dict with the polygon coordinates as an (Nx3) XYZ array
        """
        element = self[element_id]
        assert self.is_crosswalk(element)
        traffic_element = element.element.traffic_control_element

        xyz = self.unpack_deltas_cm(
            traffic_element.points_x_deltas_cm,
            traffic_element.points_y_deltas_cm,
            traffic_element.points_z_deltas_cm,
            traffic_element.geo_frame,
        )

        return {"xyz": xyz}

    def is_traffic_light(self, element_id: str) -> bool:
        """
        Check if the element is a traffic light
        Args:
            element_id (str): the id (utf-8 encode) of the element

        Returns:
            True if the element is a traffic light
        """
        element = self[element_id]
        if not element.element.HasField("traffic_control_element"):
            return False
        traffic_el = element.element.traffic_control_element
        return traffic_el.HasField("traffic_light") is True

    @lru_cache(maxsize=CACHE_SIZE)
    def is_traffic_face(self, element_id: str) -> bool:
        """
        Check if the element is a traffic light face (of any color)

        Args:
            element_id (str): the id (utf-8 encode) of the element
        Returns:
            True if the element is a traffic light face, False otherwise
        """

        for color in TLFacesColors:
            color_name = color.name
            if self.is_traffic_face_color(element_id, color_name.lower()):
                return True
        return False

    def is_traffic_face_color(self, element_id: str, color: str) -> bool:
        """
        Check if the element is a traffic light face of the given color

        Args:
            element_id (str): the id (utf-8 encode) of the element
            color (str): the color to check
        Returns:
            True if the element is a traffic light face with the given color
        """
        element = self[element_id]
        if not element.element.HasField("traffic_control_element"):
            return False
        traffic_el = element.element.traffic_control_element
        if (
                traffic_el.HasField(f"signal_{color}_face")
                or traffic_el.HasField(f"signal_left_arrow_{color}_face")
                or traffic_el.HasField(f"signal_right_arrow_{color}_face")
                or traffic_el.HasField(f"signal_upper_left_arrow_{color}_face")
                or traffic_el.HasField(f"signal_upper_right_arrow_{color}_face")
        ):
            return True
        return False

    @lru_cache(maxsize=CACHE_SIZE)
    def get_color_for_face(self, face_id: str) -> str:
        """
        Utility function. It calls `is_traffic_face_color` for a set of colors until it gets an answer.
        If no color is found, then `face_id` is not the id of a traffic light face (and we raise ValueError).

        Args:
            face_id (str): the element id
        Returns:
            str: the color as string for this traffic face
        """
        for color in TLFacesColors:
            color_name = color.name
            if self.is_traffic_face_color(face_id, color_name.lower()):
                return color_name
        raise ValueError(f"Face {face_id} has no valid color among {TLFacesColors.__members__}")

    def get_bounds(self) -> dict:
        """
        For each elements of interest returns bounds [[min_x, min_y],[max_x, max_y]] and proto ids
        Coords are computed by the MapAPI and, as such, are in the world ref system.

        Returns:
            dict: keys are classes of elements, values are dict with `bounds` and `ids` keys
        """
        lanes_ids = []
        crosswalks_ids = []

        lanes_bounds = np.empty((0, 2, 2), dtype=np.float)  # [(X_MIN, Y_MIN), (X_MAX, Y_MAX)]
        crosswalks_bounds = np.empty((0, 2, 2), dtype=np.float)  # [(X_MIN, Y_MIN), (X_MAX, Y_MAX)]

        for element in self.elements:
            element_id = MapAPI.id_as_str(element.id)

            if self.is_lane(element):
                lane = self.get_lane_coords(element_id)
                x_min = min(np.min(lane["xyz_left"][:, 0]), np.min(lane["xyz_right"][:, 0]))
                y_min = min(np.min(lane["xyz_left"][:, 1]), np.min(lane["xyz_right"][:, 1]))
                x_max = max(np.max(lane["xyz_left"][:, 0]), np.max(lane["xyz_right"][:, 0]))
                y_max = max(np.max(lane["xyz_left"][:, 1]), np.max(lane["xyz_right"][:, 1]))

                lanes_bounds = np.append(lanes_bounds, np.asarray([[[x_min, y_min], [x_max, y_max]]]), axis=0)
                lanes_ids.append(element_id)

            if self.is_crosswalk(element):
                crosswalk = self.get_crosswalk_coords(element_id)
                x_min, y_min = np.min(crosswalk["xyz"], axis=0)[:2]
                x_max, y_max = np.max(crosswalk["xyz"], axis=0)[:2]

                crosswalks_bounds = np.append(
                    crosswalks_bounds, np.asarray([[[x_min, y_min], [x_max, y_max]]]), axis=0,
                )
                crosswalks_ids.append(element_id)

        return {
            "lanes": {"bounds": lanes_bounds, "ids": lanes_ids},
            "crosswalks": {"bounds": crosswalks_bounds, "ids": crosswalks_ids},
        }

    @no_type_check
    def __getitem__(self, item: Union[int, str, bytes]) -> MapElement:
        if isinstance(item, str):
            return self.elements[self.ids_to_el[item]]
        elif isinstance(item, int):
            return self.elements[item]
        elif isinstance(item, bytes):
            return self.elements[self.ids_to_el[item.decode(ENCODING)]]
        else:
            raise TypeError("only str, bytes and int are allowed in API __getitem__")

    def __len__(self) -> int:
        return len(self.elements)

    def __iter__(self) -> Iterator:
        for i in range(len(self)):
            yield self[i]



import bisect
import warnings
from pathlib import Path
from typing import Optional

import numpy as np
from zarr import convenience

from ..data import ChunkedDataset, get_agents_slice_from_frames, get_frames_slice_from_scenes
from ..kinematic import Perturbation
from ..rasterization import Rasterizer
from .ego import EgoDataset
from .select_agents import select_agents, TH_DISTANCE_AV, TH_EXTENT_RATIO, TH_YAW_DEGREE


# WARNING: changing these values impact the number of instances selected for both train and inference!
MIN_FRAME_HISTORY = 10  # minimum number of frames an agents must have in the past to be picked
MIN_FRAME_FUTURE = 1  # minimum number of frames an agents must have in the future to be picked


class AgentDataset(EgoDataset):
    def __init__(
            self,
            cfg: dict,
            zarr_dataset: ChunkedDataset,
            rasterizer: Rasterizer,
            perturbation: Optional[Perturbation] = None,
            agents_mask: Optional[np.ndarray] = None,
            min_frame_history: int = MIN_FRAME_HISTORY,
            min_frame_future: int = MIN_FRAME_FUTURE,
    ):
        assert perturbation is None, "AgentDataset does not support perturbation (yet)"

        super(AgentDataset, self).__init__(cfg, zarr_dataset, rasterizer, perturbation)
        if agents_mask is None:  # if not provided try to load it from the zarr
            agents_mask = self.load_agents_mask()
            past_mask = agents_mask[:, 0] >= min_frame_history
            future_mask = agents_mask[:, 1] >= min_frame_future
            agents_mask = past_mask * future_mask

            if min_frame_history != MIN_FRAME_HISTORY:
                warnings.warn(
                    f"you're running with custom min_frame_history of {min_frame_history}",
                    RuntimeWarning,
                    stacklevel=2,
                )
            if min_frame_future != MIN_FRAME_FUTURE:
                warnings.warn(
                    f"you're running with custom min_frame_future of {min_frame_future}", RuntimeWarning, stacklevel=2
                )
        else:
            warnings.warn("you're running with a custom agents_mask", RuntimeWarning, stacklevel=2)

        # store the valid agents indices (N_valid_agents,)
        self.agents_indices = np.nonzero(agents_mask)[0]

        # store an array where valid indices have increasing numbers and the rest is -1 (N_total_agents,)
        self.mask_indices = agents_mask.copy().astype(np.int)
        self.mask_indices[self.mask_indices == 0] = -1
        self.mask_indices[self.mask_indices == 1] = np.arange(0, np.sum(agents_mask))

        # this will be used to get the frame idx from the agent idx
        self.cumulative_sizes_agents = self.dataset.frames["agent_index_interval"][:, 1]
        self.agents_mask = agents_mask

    def load_agents_mask(self) -> np.ndarray:
        """
        Loads a boolean mask of the agent availability stored into the zarr. Performs some sanity check against cfg.
        Returns: a boolean mask of the same length of the dataset agents

        """
        agent_prob = self.cfg["raster_params"]["filter_agents_threshold"]

        agents_mask_path = Path(self.dataset.path) / f"agents_mask/{agent_prob}"
        if not agents_mask_path.exists():  # don't check in root but check for the path
            warnings.warn(
                f"cannot find the right config in {self.dataset.path},\n"
                f"your cfg has loaded filter_agents_threshold={agent_prob};\n"
                "but that value doesn't have a match among the agents_mask in the zarr\n"
                "Mask will now be generated for that parameter.",
                RuntimeWarning,
                stacklevel=2,
            )

            select_agents(
                self.dataset,
                agent_prob,
                th_yaw_degree=TH_YAW_DEGREE,
                th_extent_ratio=TH_EXTENT_RATIO,
                th_distance_av=TH_DISTANCE_AV,
            )

        agents_mask = convenience.load(str(agents_mask_path))  # note (lberg): this doesn't update root
        return agents_mask

    def __len__(self) -> int:
        """
        length of the available and reliable agents (filtered using the mask)
        Returns: the length of the dataset

        """
        return len(self.agents_indices)

    def __getitem__(self, index: int) -> dict:
        """
        Differs from parent by iterating on agents and not AV.
        """
        if index < 0:
            if -index > len(self):
                raise ValueError("absolute value of index should not exceed dataset length")
            index = len(self) + index

        index = self.agents_indices[index]
        track_id = self.dataset.agents[index]["track_id"]
        frame_index = bisect.bisect_right(self.cumulative_sizes_agents, index)
        scene_index = bisect.bisect_right(self.cumulative_sizes, frame_index)

        if scene_index == 0:
            state_index = frame_index
        else:
            state_index = frame_index - self.cumulative_sizes[scene_index - 1]
        return self.get_frame(scene_index, state_index, track_id=track_id)

    def get_scene_dataset(self, scene_index: int) -> "AgentDataset":
        """
        Differs from parent only in the return type.
        Instead of doing everything from scratch, we rely on super call and fix the agents_mask
        """

        new_dataset = super(AgentDataset, self).get_scene_dataset(scene_index).dataset

        # filter agents_bool values
        frame_interval = self.dataset.scenes[scene_index]["frame_index_interval"]
        # ASSUMPTION: all agents_index are consecutive
        start_index = self.dataset.frames[frame_interval[0]]["agent_index_interval"][0]
        end_index = self.dataset.frames[frame_interval[1] - 1]["agent_index_interval"][1]
        agents_mask = self.agents_mask[start_index:end_index].copy()

        return AgentDataset(
            self.cfg, new_dataset, self.rasterizer, self.perturbation, agents_mask  # overwrite the loaded one
        )

    def get_scene_indices(self, scene_idx: int) -> np.ndarray:
        """
        Get indices for the given scene. Here __getitem__ iterate over valid agents indices.
        This means ``__getitem__(0)`` matches the first valid agent in the dataset.
        Args:
            scene_idx (int): index of the scene

        Returns:
            np.ndarray: indices that can be used for indexing with __getitem__
        """
        scenes = self.dataset.scenes
        assert scene_idx < len(scenes), f"scene_idx {scene_idx} is over len {len(scenes)}"
        frame_slice = get_frames_slice_from_scenes(scenes[scene_idx])
        agent_slice = get_agents_slice_from_frames(*self.dataset.frames[frame_slice][[0, -1]])

        mask_valid_indices = (self.agents_indices >= agent_slice.start) * (self.agents_indices < agent_slice.stop)
        indices = np.nonzero(mask_valid_indices)[0]
        return indices

    def get_frame_indices(self, frame_idx: int) -> np.ndarray:
        """
        Get indices for the given frame. Here __getitem__ iterate over valid agents indices.
        This means ``__getitem__(0)`` matches the first valid agent in the dataset.
        Args:
            frame_idx (int): index of the scene

        Returns:
            np.ndarray: indices that can be used for indexing with __getitem__
        """
        assert frame_idx < len(self.dataset.frames), f"frame_idx {frame_idx} is over len {len(self.dataset.frames)}"

        # avoid using `get_agents_slice_from_frames` as it hits the disk
        agent_start = self.cumulative_sizes_agents[frame_idx - 1] if frame_idx > 0 else 0
        agent_end = self.cumulative_sizes_agents[frame_idx]
        # slice using frame boundaries and take only valid indices
        mask_idx = self.mask_indices[agent_start:agent_end]
        indices = mask_idx[mask_idx != -1]
        return indices



from typing import Iterable, Tuple

import cv2
import imageio
import numpy as np


def write_gif(
        output_filepath: str,
        images: Iterable[np.ndarray],
        resolution: Tuple[int, int],
        fps: float = 24.0,
        loop: int = 0,
        interpolation: int = cv2.INTER_CUBIC,
) -> None:
    """Writes input RGB images to given output gif filepath using imageio. It resizes images
    if necessary using given interpolation (default=``cv2.INTER_CUBIC``).

    Arguments:
        output_filepath (str): output filepath, should end in .gif
        images (Iterable[np.ndarray]): a list or other iterable of images.
        resolution (Tuple[int, int]): desired resolution, e.g. (512, 512)

    Keyword Arguments:
        fps (float): Frames per second (default: {24.0})
        loop (int): 0 means loop forever, any other number loops the GIF that many times (default: {0})
        interpolation (int): Interpolation to be used when resizing (default: {cv2.INTER_CUBIC})
    """

    duration = 1 / fps
    resized_images = []

    for img in images:
        # Go from C,0,1 to 0,1,C ordering
        if len(img.shape) == 3 and img.shape[0] == 3:
            img = img.transpose(1, 2, 0)

        if img.shape[:2] != resolution:
            img = cv2.resize(img, resolution, interpolation=interpolation)
        resized_images.append(img)

    imageio.mimwrite(output_filepath, resized_images, duration=duration, loop=loop)
